# Copyright 2025 IBM Corp.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import torch
from .vqvae import VQ


class VQBackbone(VQ):
    def __init__(self, return_type: str = "embedding", **kwargs):
        """
        Backbone class for VQ models.

        return_type: "embedding", "latent", or "quantized"
        """
        super().__init__(**kwargs)
        self.return_type = return_type
        if self.return_type == "embedding":
            self.out_channels = [self.enc_dim] * len(self.encoder.blocks)
        elif self.return_type in ["latent", "quantized"]:
            self.out_channels = [self.latent_dim]
        else:
            raise NotImplementedError

    def forward(self, x: torch.Tensor, **kwargs) -> torch.Tensor:
        """Forward pass of the encoder and quantizer.

        Args:
            x: Input image tensor of shape B C H W
              or B H W in case of semantic segmentation

        Returns:
            Embedding, latent, or quantized latent code of shape B D_Q H_Q W_Q
        """
        x = self.prepare_input(x)

        embedding = self.encoder(x, return_intermediates=True)
        latent = self.quant_proj(embedding[-1])

        if self.return_type == "embedding":
            return embedding
        elif self.return_type == "latent":
            return latent
        elif self.return_type == "quantized":
            quant, code_loss, tokens = self.quantize(latent)
            return quant




import warnings
import torch
import torch.nn as nn
from torchvision.models.detection.transform import GeneralizedRCNNTransform
from torchvision.models.detection.image_list import ImageList
from terratorch.registry import BACKBONE_REGISTRY

class DecoderNotFoundError(Exception):
    pass

def extract_prefix_keys(d: dict, prefix: str) -> dict:
    extracted_dict = {}
    remaining_dict = {}
    for k, v in d.items():
        if k.startswith(prefix):
            extracted_dict[k[len(prefix) :]] = v
        else:
            remaining_dict[k] = v

    return extracted_dict, remaining_dict


def get_image_size(x):
    if isinstance(x, torch.Tensor):
        return x.shape[-2:]
    elif isinstance(x, dict):
        # Multimodal input
        shapes = []
        for modality, tensor in x.items():
            if isinstance(tensor, torch.Tensor) and tensor.ndim in (4, 5):
                shapes.append(tensor.shape[-2:])
        if not shapes:
            raise ValueError(f"No image modality among inputs {list(x.keys())}. "
                             f"TerraTorch expects image inputs in shapes: [B, C, H, W] or [B, C, T, H, W]")
        if len(set(shapes)) != 1:
            warnings.warn(f"Modalities have different spatial shapes: {shapes}. Using largest shape.")
            # Sort by area and pick largest
            shapes.sort(key=lambda s: s[0]*s[1], reverse=True)
        return shapes[0]
    else:
        raise ValueError('Could not infer image shape. Expecting tensor or dict of tensors.')

def center_crop(x: torch.Tensor, out_size: tuple[int, int]) -> torch.Tensor:
    h_out, w_out = out_size
    h, w = x.shape[-2], x.shape[-1]
    top = max((h - h_out) // 2, 0)
    left = max((w - w_out) // 2, 0)
    return x[..., top:top + h_out, left:left + w_out]

def pad_images(imgs: torch.Tensor | dict[str, torch.Tensor], patch_size: int | list, padding: str
               ) -> torch.Tensor | dict[str, torch.Tensor]:
    if isinstance(imgs, dict):
        # Handle multimodal data
        return {modality: pad_images(mod_imgs, patch_size, padding) for modality, mod_imgs in imgs.items()}
    elif not isinstance(imgs, torch.Tensor):
        # Expect tensor
        return imgs
    elif imgs.ndim < 4:
        # Only pad images with format [B, C, T, H, W] or [B, C, T, H, W]
        return imgs

    p_t = 1
    if isinstance(patch_size, int):
         p_h = p_w = patch_size
    elif len(patch_size) == 1:
        p_h = p_w = patch_size[0]
    elif len(patch_size) == 2:
        p_h, p_w = patch_size
    elif len(patch_size) == 3:
        p_t, p_h, p_w = patch_size
    else:
        raise ValueError(f'patch size {patch_size} not valid, must be int or list of ints with length 1, 2, or 3.')

    # Double the patch size to ensure the resulting number of patches is divisible by 2 (required for many decoders)
    p_h, p_w = p_h * 2, p_w * 2

    if p_t > 1 and imgs.ndim < 5:
        raise ValueError(f"Multi-temporal padding requested (p_t = {p_t}) "
                         f"but no multi-temporal data provided (data shape = {imgs.shape})."
                         f"Expecting tensor or dict of tensors with shape [B, C, T, H, W].")

    h, w = imgs.shape[-2:]
    t = imgs.shape[-3] if imgs.ndim > 4 else 1
    t_pad, h_pad, w_pad = (p_t - t % p_t) % p_t, (p_h - h % p_h) % p_h, (p_w - w % p_w) % p_w
    # Split padding equally on both sides
    h_pad_left, h_pad_right = h_pad // 2, h_pad - h_pad // 2
    w_pad_left, w_pad_right = w_pad // 2, w_pad - w_pad // 2

    if t_pad > 0:
        # Multi-temporal padding
        imgs = torch.stack([
            nn.functional.pad(img, (w_pad_left, w_pad_right, h_pad_left, h_pad_right, 0, t_pad), mode=padding)
            for img in imgs
        ])
    elif h_pad > 0 or w_pad > 0:
        imgs = torch.stack([
            nn.functional.pad(img, (w_pad_left, w_pad_right, h_pad_left, h_pad_right), mode=padding)
            for img in imgs
        ])
    return imgs

def _get_backbone(backbone: str | nn.Module, **backbone_kwargs) -> nn.Module:
    use_temporal = backbone_kwargs.pop('use_temporal', False)
    pooling = backbone_kwargs.pop('temporal_pooling', 'mean')
    concat = backbone_kwargs.pop('temporal_concat', None)
    n_timestamps = backbone_kwargs.pop('temporal_n_timestamps', None)
    features_permute_op = backbone_kwargs.pop('temporal_features_permute_op', None)
    subset_lengths = backbone_kwargs.pop('temporal_subset_lengths', None)

    if isinstance(backbone, nn.Module):
        model = backbone
    else:
        model = BACKBONE_REGISTRY.build(backbone, **backbone_kwargs)

    # Apply TemporalWrapper inside _get_backbone
    if use_temporal:
        model = TemporalWrapper(model, pooling=pooling, concat=concat, n_timestamps=n_timestamps,
                                features_permute_op=features_permute_op, subset_lengths=subset_lengths)

    return model

class TemporalWrapper(nn.Module):
    def __init__(self, encoder: nn.Module, pooling: str = 'mean', concat: bool = None, n_timestamps: int | None = None,
                 features_permute_op: list[int]| None = None, subset_lengths: list[int]| None = None):
        """
        Wrapper for applying a temporal encoder across multiple time steps.

        Args:
            encoder (nn.Module): The feature extractor (backbone).
            pooling (str): Type of pooling ('keep', 'concat', 'mean', 'max', 'diff'), 'diff' requires exactly two temporal elements after optional subset aggregation.
            concat (bool): Deprecated - 'concat' now integrated as pooling option.
            n_timestamps (int): Used to compute backbone out_channels when constructing encoder–decoder pipelines in case of 'concat' pooling.
            features_permute_op (list): Permutation operation to perform on the features before aggregation.
                This is in case the features to do not match either 'BCHW' or 'BLC' formats. It is reversed once
                aggregation has happened.
            subset_lengths (list[int], optional): If set, performs two-step temporal aggregation: 
                (1) split timesteps into defined subsets (e.g., [2, 3] → first 2 and last 3 timesteps),
                average within each; (2) apply the selected `pooling` across the resulting subset means. 
                Lengths must match the total number of timesteps. For pooling='diff', exactly two subsets are required.
        """
        super().__init__()

        # Warn if deprecated args are used
        if concat is not None:
            warnings.warn(
                "'concat' is deprecated in TemporalWrapper. "
                "Use pooling='concat' instead.",
                DeprecationWarning
                )

        # Check supported pooling modes
        supported_poolings = ["mean", "max", "diff", "keep", "concat"]
        if pooling not in supported_poolings:
            raise ValueError(f"Unsupported pooling '{pooling}', choose from {supported_poolings}.")
        
        self.encoder = encoder
        self.pooling = pooling
        self.features_permute_op = features_permute_op
        self.subset_lengths = subset_lengths

        # Validate subset_lengths defines two subsets for diff pooling and matches n_timestamps if set
        if subset_lengths is not None:
            if pooling == "diff" and len(subset_lengths) != 2:
                raise ValueError(
                    f"`diff` pooling requires exactly two subsets in `subset_lengths`, "
                    f"but got {len(subset_lengths)}: {subset_lengths}"
                )

            if n_timestamps is not None and sum(subset_lengths) != n_timestamps:
                raise ValueError(
                    f"The sum of `subset_lengths` must equal `n_timestamps` "
                    f"(got sum={sum(subset_lengths)}, n_timestamps={n_timestamps})."
                )

        # Precompute reverse permutation for restoring original dims after processing
        if features_permute_op is not None:
            self.reverse_permute_op = [None] * len(features_permute_op)
            for i, p in enumerate(features_permute_op):
                self.reverse_permute_op[p] = i
        else:
            self.reverse_permute_op = None

        if hasattr(encoder, "out_channels"):
            if pooling == "concat":
                if n_timestamps is None:
                    warnings.warn(
                        "Cannot derive `out_channels` for 'concat' pooling without `n_timestamps`"
                        "(Required to build Encoder–Decoder models).",
                        UserWarning,
                    )
                    self.out_channels = None
                else:
                    self.out_channels = [c * n_timestamps for c in encoder.out_channels]
            else:
                self.out_channels = encoder.out_channels
    
    
    def pool_temporal(self, stacked: torch.Tensor, pooling: str, subset_lengths: list[int] | None):
        """
        Pool per-timestep outputs based on the specified pooling method.
        """

        T = stacked.shape[1]  # number of timesteps

        if subset_lengths is not None:
            if sum(subset_lengths) != T:
                raise ValueError(
                    f"Sum of `subset_lengths` ({sum(subset_lengths)}) must equal timesteps ({T})."
                )

            # Split into subsets and average within each
            subset_splits = torch.split(stacked, subset_lengths, dim=1)
            stacked = torch.stack(
                [subset.mean(dim=1) for subset in subset_splits],
                dim=1
            )

        if pooling == "concat":
            return stacked.flatten(1, 2) # concat over time: [B, T*C, H, W]
        
        elif pooling == "max":
            return torch.max(stacked, dim=1).values # max over time: [B, C, H, W]
        
        elif pooling == "diff":
            if stacked.shape[1] != 2:
                raise ValueError(
                    f"`diff` pooling requires exactly two temporal elements "
                    f"(from input or after subset aggregation), got {stacked.shape[1]}. "
                    f"Consider to use `subset_lengths` to define two subsets."
                )
            return stacked[:, 0] - stacked[:, 1] # difference between two timesteps: [B, C, H, W]
        
        elif pooling == "mean":
            return torch.mean(stacked, dim=1) # mean over time: [B, C, H, W]
        
        else: 
            return stacked # "keep": return [B, T, C, H, W] sequence
        
    def reshape_5d(self, tensor: torch.Tensor, B: int, T: int) -> torch.Tensor:
        """
        Reshape latent to [B, T, C, ...]. Appends dummy dim for ViTs.
        """
        if tensor.dim() == 4:  # Conv backbone: [BT, C, H, W]
            C, H, W = tensor.shape[1:]
            return tensor.reshape(B, T, C, H, W)
        elif tensor.dim() == 3:  # ViT backbone: [BT, L, C]
            L, C = tensor.shape[-2:]
            x = tensor.reshape(B, T, L, C)
            return x.permute(0, 1, 3, 2).unsqueeze(-1)
        raise ValueError(f"Expected latent tensor to be 3D or 4D, but got {tensor.dim()}.")
    
    def vit_postprocess(self, tensor: torch.Tensor) -> torch.Tensor:
        """
        Reorders L and C dim and removes helper size-1 dim if present.
        """
        if tensor.shape[-1] != 1:
            return tensor
        return tensor.squeeze(-1).movedim(-1, -2)
    
    def permute_op(self, tensor: torch.Tensor, permute_op: list[int]) -> torch.Tensor:
        """
        Apply a permutation operation to the tensor.
        """
        if permute_op is not None:
            if len(permute_op) != len(tensor.shape):
                raise ValueError(f"Expected permute_op to have same number of dimensions as tensor, but got {len(permute_op)} and {len(tensor.shape)}")
            return torch.permute(tensor, permute_op)
        return tensor
        
    def forward(self, 
                x: torch.Tensor | dict[str, torch.Tensor]
                ) -> list[torch.Tensor | dict[str, torch.Tensor]]:
        """
        Forward pass for temporal processing.

        Args:
            x: Input tensor of shape [B, C, T, H, W] or dict with shape {modality: [B, C_mod, T, H, W]}.
        Returns:
            List: A list of processed tensors/dicts, one per feature map.
        """

        # Handle dict input (multimodal) or single tensor
        is_dict = isinstance(x, dict)
        sample = next(iter(x.values())) if is_dict else x

        # Input must have 5 dims: [B, C, T, H, W]
        if sample.dim() != 5:
            raise ValueError(f"Expected input shape [B, C, T, H, W], got {tuple(sample.shape)}")

        # Flatten temporal dimension into batch for encoder forward pass
        if is_dict:
            B, _, T, H, W = sample.shape
            flat_input = {
                k: v.permute(0, 2, 1, 3, 4).reshape(-1, v.shape[1], *v.shape[3:])
                for k, v in x.items()
            }
        else:
            B, C, T, H, W = sample.shape
            flat_input = x.permute(0, 2, 1, 3, 4).reshape(B * T, C, H, W)

        # Run encoder backbone
        feat = self.encoder(flat_input)
        if not isinstance(feat, (list, tuple)):
            feat = [feat]
        else:
            feat = list(feat)

        outputs = []

        # Postprocess each feature map returned by encoder (layer outputs)
        for feature_map in feat:
            if isinstance(feature_map, dict): # multimodal output
                mod_keys = feature_map.keys()
                pooled = {}
                for k in mod_keys:
                    # Apply optional permutation before processing
                    feature_map[k] = self.permute_op(feature_map[k], self.features_permute_op)

                    # Reshape back to [B, T, ...]
                    mod_stacked = self.reshape_5d(feature_map[k], B, T)

                    # Temporal pooling
                    pooled[k] = self.pool_temporal(mod_stacked, self.pooling, self.subset_lengths)

                    # Reverse permutation to restore original dim order
                    pooled[k] = self.permute_op(pooled[k], self.reverse_permute_op)

                    # ViT postprocessing if needed
                    if pooled[k].shape[-1] == 1: 
                        pooled[k] = self.vit_postprocess(pooled[k])

            else: # single-modality feature map
                feature_map = self.permute_op(feature_map, self.features_permute_op)
                stacked = self.reshape_5d(feature_map, B, T)
                pooled = self.pool_temporal(stacked, self.pooling, self.subset_lengths)
                pooled = self.permute_op(pooled, self.reverse_permute_op)

                if pooled.shape[-1] == 1: 
                    pooled = self.vit_postprocess(pooled)
                    
            outputs.append(pooled)

        return outputs

class TerratorchGeneralizedRCNNTransform(GeneralizedRCNNTransform):
    
    def init(min_size: int,
             max_size: int,
             image_mean: list[float],
             image_std: list[float],
             size_divisible: int = 32,
             fixed_size: tuple[int, int] | None = None,
             **kwargs):
        
        super().__init__(min_size,
                         max_size,
                         image_mean,
                         image_std,
                         size_divisible,
                         fixed_size,
                         **kwargs)
        
    def forward(
        self, images: list[torch.Tensor], targets: list[dict[str, torch.Tensor]] | None = None
    ) -> tuple[ImageList, list[dict[str, torch.Tensor]]] | None:
        images = [img for img in images]
        if targets is not None:
            # make a copy of targets to avoid modifying it in-place
            # once torchscript supports dict comprehension
            # this can be simplified as follows
            # targets = [{k: v for k,v in t.items()} for t in targets]
            targets_copy: list[dict[str, torch.Tensor]] = []
            for t in targets:
                data: dict[str, torch.Tensor] = {}
                for k, v in t.items():
                    data[k] = v
                targets_copy.append(data)
            targets = targets_copy
        for i in range(len(images)):
            image = images[i]
            target_index = targets[i] if targets is not None else None
            images[i] = image
            if targets is not None and target_index is not None:
                targets[i] = target_index
        image_sizes = [img.shape[-2:] for img in images]
        images = [x[None] for x in images]
        images = torch.cat(images, 0)
        image_sizes_list: list[tuple[int, int]] = []
        for image_size in image_sizes:
            torch._assert(
                len(image_size) == 2,
                f"Input tensors expected to have in the last two elements H and W, instead got {image_size}",
            )
            image_sizes_list.append((image_size[0], image_size[1]))

        image_list = ImageList(images, image_sizes_list)
        return image_list, targets


from torchgeo.datasets.utils import (
    Path,
    check_integrity,
    download_and_extract_archive,
    download_url,
    lazy_import,
    percentile_normalization,
)

from collections.abc import Callable
from typing import Any, ClassVar

from torch import Tensor
import matplotlib.pyplot as plt
from matplotlib.figure import Figure
from matplotlib import patches
from functools import partial

# from terratorch.datasets import mVHR10
from terratorch.datasets.substation import Substation

from torchgeo.datamodules import NonGeoDataModule

import albumentations as A
from albumentations.pytorch import transforms as T
import torchvision.transforms as orig_transforms

from torch.utils.data import DataLoader

import torch
from torch import nn
import numpy as np
import terratorch

import pdb

def collate_fn_detection(batch):
    new_batch = {
        "image": [item["image"] for item in batch],
        "boxes": [item["boxes"] for item in batch],
        "labels": [item["labels"] for item in batch],
        "masks": [item["masks"] for item in batch],
    }
    return new_batch


def get_transform(train, image_size=228, n_timesteps=4):
    transforms = []
    if image_size != 228:
        transforms.append(A.Resize(height=image_size, width=image_size))
    if n_timesteps > 1:
        transforms.append(terratorch.datasets.transforms.FlattenTemporalIntoChannels())
    transforms.append(T.ToTensorV2())
    if n_timesteps > 1:
        transforms.append(terratorch.datasets.transforms.UnflattenTemporalFromChannels(n_timesteps=n_timesteps))
    return A.Compose(transforms, bbox_params=A.BboxParams(format="pascal_voc", label_fields=['labels']), is_check_shapes=False)


def apply_transforms(sample, transforms):


    # pdb.set_trace()
    sample['image'] = torch.stack(tuple(sample["image"]))
    sample['image'] = sample['image'].permute(1, 2, 0) if len(sample['image'].shape) == 3 else sample['image'].permute(0, 2, 3, 1)
    sample['image'] = np.array(sample['image'].cpu())
    sample["masks"] = [np.array(torch.stack(tuple(x)).cpu()) for x in sample["masks"]]
    sample["boxes"] = np.array(sample["boxes"].cpu())
    sample["labels"] = np.array(sample["labels"].cpu())
    transformed = transforms(image=sample['image'],
                             masks=sample["masks"], 
                             bboxes=sample["boxes"],
                             labels=sample["labels"])

    indexes = [i for i, x in enumerate(transformed['masks']) if x.any()]
    transformed['masks'] = [x for i, x in enumerate(transformed['masks']) if i in indexes]
    transformed['boxes'] = torch.tensor(transformed['bboxes'], dtype=torch.float32)
    transformed['boxes'] = transformed['boxes'][indexes]
    transformed['labels'] = torch.tensor(transformed['labels'], dtype=torch.int64)[indexes]

    del transformed['bboxes']

    return transformed


class Normalize(Callable):
    def __init__(self, means, stds, max_pixel_value=None):
        super().__init__()
        self.means = means
        self.stds = stds
        self.max_pixel_value = max_pixel_value

    def __call__(self, batch):

        batch['image']=torch.stack(tuple(batch["image"]))
        image = batch["image"]/self.max_pixel_value if self.max_pixel_value is not None else batch["image"]
        if len(image.shape) == 5:
            means = self.means.clone().detach().to(device=image.device).view(1, -1, 1, 1, 1)
            stds = self.stds.clone().detach().to(device=image.device).view(1, -1, 1, 1, 1)
        elif len(image.shape) == 4:
            means = self.means.clone().detach().to(device=image.device).view(1, -1, 1, 1)
            stds = self.stds.clone().detach().to(device=image.device).view(1, -1, 1, 1)
        else:
            msg = f"Expected batch to have 5 or 4 dimensions, but got {len(image.shape)}"
            raise Exception(msg)
        batch["image"] = (image - means) / stds
        # pdb.set_trace()
        return batch


class IdentityTransform(nn.Module):
    def __init__(self):
        super().__init__()

    def forward(self, x):
        return x

means = [1423.3745710003543, 1219.2943391189588, 1182.4678074598091, 1158.0459084351446, 1401.8725467408829, 
         2198.8962678953585, 2582.976482945787, 2505.9577635634005, 2812.9741141745803, 808.0196180351925, 
         15.818041869437652, 2185.9854710710974, 1469.6518299083612]


stds = [285.46691956407517, 394.44991388670013, 450.0629924473796, 671.2598383733988, 594.6671241798275,
        618.4261097309303, 787.2213184558726, 795.4738600694346, 862.1076233925734, 331.81864168987255, 
        14.444442475684806, 840.1079725056068, 818.1891190853091]


class SubstationDataModule(NonGeoDataModule):
    def __init__(
        self,
        root: Path = "/opt/app-root/src/use-cases/substation-dataset-conversion/Substation",
        bands=[3, 2, 1],
        mask_2d: bool = False,
        batch_size: int = 8,
        num_workers: int = 4, 
        means=means,
        stds=stds,
        timepoint_aggregation: str = 'concat',
        download: bool = False,
        checksum: bool = False,
        num_of_timepoints: int = 4,
        use_timepoints: bool = True,
        mode: str = 'segmentation',
        collate_fn=None,
        dataset_version: str = 'full',
        image_size: int = 224,
        train_transform=None,
        val_transform=None,
        test_transform=None,
        plot_indexes = [2,1,0],
        *args,
        **kwargs):

        super().__init__(SubstationDataModule,
                         batch_size=batch_size,
                         num_workers=num_workers,
                         root=root, 
                         bands=bands,
                         mask_2d=mask_2d,
                         timepoint_aggregation=timepoint_aggregation,
                         download=download,
                         checksum=checksum,
                         num_of_timepoints=num_of_timepoints,
                         use_timepoints=use_timepoints,
                         mode=mode,
                         dataset_version=dataset_version,
                         train_transform=train_transform,
                         val_transform=val_transform,
                         test_transform=test_transform,
                         **kwargs)

        if (use_timepoints == False) | (timepoint_aggregation=='median'):
            num_of_timepoints = 1
        if (mode == 'object_detection'):
            self.train_transform = partial(apply_transforms,transforms=get_transform(True, image_size, num_of_timepoints)) if train_transform is None else train_transform
            self.val_transform = partial(apply_transforms,transforms=get_transform(False, image_size, num_of_timepoints)) if val_transform is None else val_transform
            self.test_transform = partial(apply_transforms,transforms=get_transform(False, image_size, num_of_timepoints)) if test_transform is None else test_transform
        else:
            self.train_transform = train_transform
            self.val_transform = val_transform
            self.test_transform = test_transform
        
        self.bands = bands
        self.means = torch.tensor(means)[self.bands]
        self.stds = torch.tensor(stds)[self.bands]
        self.root = root
        self.aug = Normalize(self.means, self.stds)      
        self.mask_2d = mask_2d
        self.timepoint_aggregation = timepoint_aggregation
        self.download = download
        self.checksum = checksum
        self.collate_fn = collate_fn_detection if collate_fn is None else collate_fn
        self.num_of_timepoints = num_of_timepoints
        self.use_timepoints = use_timepoints
        self.mode = mode
        self.dataset_version = dataset_version
        self.image_size = image_size
        self.plot_indexes = plot_indexes

    def setup(self, stage: str) -> None:
        
        if stage in ["fit"]:
            self.train_dataset = Substation(
                        root = self.root,
                        bands = self.bands,
                        mask_2d = self.mask_2d,
                        timepoint_aggregation = self.timepoint_aggregation,
                        download = self.download,
                        checksum = self.checksum,
                        transforms = self.train_transform,
                        num_of_timepoints = self.num_of_timepoints,
                        use_timepoints = self.use_timepoints,
                        mode = self.mode,
                        split = "train",
                        dataset_version = self.dataset_version,
                        plot_indexes = self.plot_indexes
                        )            
        if stage in ["fit", "validate"]:
            self.val_dataset = Substation(
                        root = self.root,
                        bands = self.bands,
                        mask_2d = self.mask_2d,
                        timepoint_aggregation = self.timepoint_aggregation,
                        download = self.download,
                        checksum = self.checksum,
                        transforms = self.train_transform,
                        num_of_timepoints = self.num_of_timepoints,
                        use_timepoints = self.use_timepoints,
                        mode = self.mode,
                        split = "val",
                        dataset_version = self.dataset_version,
                        plot_indexes = self.plot_indexes
            )
        if stage in ["test"]:
            self.test_dataset = Substation(
                        root = self.root,
                        bands = self.bands,
                        mask_2d = self.mask_2d,
                        timepoint_aggregation = self.timepoint_aggregation,
                        download = self.download,
                        checksum = self.checksum,
                        transforms = self.train_transform,
                        num_of_timepoints = self.num_of_timepoints,
                        use_timepoints = self.use_timepoints,
                        mode = self.mode,
                        split = "test",
                        dataset_version = self.dataset_version,
                        plot_indexes = self.plot_indexes
            )
        

    def _dataloader_factory(self, split: str) -> DataLoader[dict[str, Tensor]]:
        """Implement one or more PyTorch DataLoaders.

        Args:
            split: Either 'train', 'val', 'test', or 'predict'.

        Returns:
            A collection of data loaders specifying samples.

        Raises:
            MisconfigurationException: If :meth:`setup` does not define a
                dataset or sampler, or if the dataset or sampler has length 0.
        """
        dataset = self._valid_attribute(f"{split}_dataset", "dataset")
        batch_size = self.batch_size

        return DataLoader(
            dataset=dataset,
            batch_size=batch_size,
            shuffle=split == "train",
            num_workers=self.num_workers,
            collate_fn=self.collate_fn
        )





import gc
import pdb

import pytest
import torch

from terratorch.tasks.object_detection_task import ObjectDetectionTask


def task1():
    return ObjectDetectionTask(
        model_factory="ObjectDetectionModelFactory",
        model_args={
            "framework": "faster-rcnn",
            "backbone": "prithvi_eo_v2_300",
            "num_classes": 12,
            "backbone_pretrained": False,
            "backbone_bands": ["RED", "GREEN", "BLUE"],
            "necks": [
                {"name": "SelectIndices", "indices": [5, 11, 17, 23]},
                {"name": "ReshapeTokensToImage"},
                {"name": "LearnedInterpolateToPyramidal"},
                {"name": "FeaturePyramidNetworkNeck"},
            ],
        },
        lr=0.001,
        optimizer="Adam",
        optimizer_hparams={},
        scheduler=None,
        scheduler_hparams={},
        freeze_backbone=False,
        freeze_decoder=False,
        class_names=None,
        iou_threshold=0.5,
        score_threshold=0.5,
        boxes_field="boxes",
        labels_field="labels",
        masks_field="masks",
    )


def task2():
    return ObjectDetectionTask(
        model_factory="ObjectDetectionModelFactory",
        model_args={
            "framework": "faster-rcnn",
            "backbone": "timm_resnet50",
            "backbone_pretrained": False,
            "num_classes": 12,
            "in_channels": 3,
            "necks": [{"name": "FeaturePyramidNetworkNeck"}],
        },
        lr=0.001,
        optimizer="Adam",
        optimizer_hparams={},
        scheduler=None,
        scheduler_hparams={},
        freeze_backbone=False,
        freeze_decoder=False,
        class_names=None,
        iou_threshold=0.5,
        score_threshold=0.5,
        boxes_field="bbox_xyxy",
        labels_field="label",
        masks_field="mask",
    )


def dummy_batch1():
    return {
        "image": torch.randn(2, 3, 256, 256),
        "boxes": [
            torch.tensor([[10, 10, 50, 50], [20, 20, 60, 60]], dtype=torch.float32),
            torch.tensor([[20, 20, 60, 60]], dtype=torch.float32),
        ],
        "labels": [torch.tensor([1, 2]), torch.tensor([2])],
    }


def dummy_batch2():
    return {
        "image": torch.randn(2, 3, 256, 256),
        "bbox_xyxy": [
            torch.tensor([[10, 10, 50, 50], [20, 20, 60, 60]], dtype=torch.float32),
            torch.tensor([[20, 20, 60, 60]], dtype=torch.float32),
        ],
        "label": [torch.tensor([1, 2]), torch.tensor([2])],
    }


tasks = [task1(), task2()]

dummy_batches = [dummy_batch1(), dummy_batch2()]


@pytest.mark.parametrize("task", tasks)
def test_initialization(task):
    assert hasattr(task, "model_factory")
    assert hasattr(task, "monitor")
    assert task.score_threshold == 0.5


@pytest.mark.parametrize("task", tasks)
def test_configure_models(task):
    task.configure_models()
    assert hasattr(task, "model")
    assert callable(task.model.forward)


@pytest.mark.parametrize("task", tasks)
def test_configure_metrics(task):
    task.configure_metrics()
    assert hasattr(task, "train_metrics")
    assert hasattr(task, "val_metrics")
    assert hasattr(task, "test_metrics")


@pytest.mark.parametrize("task", tasks)
def test_configure_optimizers(task):
    opt = task.configure_optimizers()
    assert isinstance(opt, dict) or hasattr(opt, "optimizer")


@pytest.mark.parametrize("task,dummy_batch", zip(tasks, dummy_batches, strict=False))
def test_reformat_batch(task, dummy_batch):
    batch_size = 2
    reformatted = task.reformat_batch(dummy_batch, batch_size)
    assert isinstance(reformatted, list)
    assert "boxes" in reformatted[0]
    assert "labels" in reformatted[0]


@pytest.mark.parametrize("task,dummy_batch", zip(tasks, dummy_batches, strict=False))
def test_apply_ignore_index(task, dummy_batch):
    task.ignore_index = 1
    batch_size = 2

    filtered = task.apply_ignore_index(dummy_batch, task.ignore_index)
    reformatted = task.reformat_batch(filtered, batch_size)

    for i in range(len(reformatted)):
        assert not (reformatted[i]["labels"] == 1).any()


@pytest.mark.parametrize("task", tasks)
def test_apply_nms_sample(task):
    sample = {
        "boxes": torch.tensor([[0, 0, 10, 10], [0, 0, 10, 10]], dtype=torch.float32),
        "scores": torch.tensor([0.9, 0.8]),
        "labels": torch.tensor([1, 1]),
    }
    filtered = task.apply_nms_sample(sample)
    assert filtered["boxes"].shape[0] <= 2
    gc.collect()


@pytest.mark.parametrize("task", tasks)
def test_apply_nms_batch(task):
    batch = [
        {
            "boxes": torch.tensor([[0, 0, 10, 10], [0, 0, 10, 10]], dtype=torch.float32),
            "scores": torch.tensor([0.9, 0.8]),
            "labels": torch.tensor([1, 1]),
        },
        {
            "boxes": torch.tensor([[5, 5, 15, 15]], dtype=torch.float32),
            "scores": torch.tensor([0.95]),
            "labels": torch.tensor([2]),
        },
    ]
    filtered = task.apply_nms_batch(batch, batch_size=2)
    assert isinstance(filtered, list)
    assert all("boxes" in pred for pred in filtered)
    gc.collect()


@pytest.mark.parametrize("task,dummy_batch", zip(tasks, dummy_batches, strict=False))
def test_training_step(task, dummy_batch):
    task.configure_models()
    loss = task.training_step(dummy_batch, batch_idx=0)
    assert isinstance(loss, torch.Tensor)
    gc.collect()


@pytest.mark.parametrize("task,dummy_batch", zip(tasks, dummy_batches, strict=False))
def test_predict_step(task, dummy_batch):
    task.configure_models()

    task.model.eval()
    predictions = task.predict_step(dummy_batch, batch_idx=0)
    assert isinstance(predictions, list)
    assert all("boxes" in pred for pred in predictions)
    gc.collect()



from typing import Any

import albumentations as A  # noqa: N812

from terratorch.datamodules.utils import wrap_in_compose_is_list
from terratorch.datasets import Sen4AgriNet
from torchgeo.datamodules import NonGeoDataModule


class Sen4AgriNetDataModule(NonGeoDataModule):
    """NonGeo LightningDataModule implementation for Sen4AgriNet."""

    def __init__(
        self,
        bands: list[str] | None = None,
        batch_size: int = 8,
        num_workers: int = 0,
        data_root: str = "./",
        train_transform: A.Compose | None | list = None,
        val_transform: A.Compose | None | list = None,
        test_transform: A.Compose | None | list = None,
        predict_transform: A.Compose | None | list = None,
        seed: int = 42,
        scenario: str = "random",
        requires_norm: bool = True,
        binary_labels: bool = False,
        linear_encoder: dict = None,
        **kwargs: Any,
    ) -> None:
        """
        Initializes the Sen4AgriNetDataModule for the Sen4AgriNet dataset.

        Args:
            bands (list[str] | None, optional): List of bands to use. Defaults to None.
            batch_size (int, optional): Batch size for DataLoaders. Defaults to 8.
            num_workers (int, optional): Number of workers for data loading. Defaults to 0.
            data_root (str, optional): Root directory of the dataset. Defaults to "./".
            train_transform (A.Compose | None | list, optional): Transformations for training data.
            val_transform (A.Compose | None | list, optional): Transformations for validation data.
            test_transform (A.Compose | None | list, optional): Transformations for test data.
            predict_transform (A.Compose | None | list, optional): Transformations for prediction data.
            seed (int, optional): Random seed for reproducibility. Defaults to 42.
            scenario (str): Defines the splitting scenario to use. Options are:
                - 'random': Random split of the data.
                - 'spatial': Split by geographical regions (Catalonia and France).
                - 'spatio-temporal': Split by region and year (France 2019 and Catalonia 2020).
            requires_norm (bool, optional): Whether normalization is required. Defaults to True.
            binary_labels (bool, optional): Whether to use binary labels. Defaults to False.
            linear_encoder (dict, optional): Mapping for label encoding. Defaults to None.
            **kwargs: Additional keyword arguments.
        """
        super().__init__(
            Sen4AgriNet,
            batch_size=batch_size,
            num_workers=num_workers,
            **kwargs,
        )
        self.bands = bands
        self.seed = seed
        self.train_transform = wrap_in_compose_is_list(train_transform)
        self.val_transform = wrap_in_compose_is_list(val_transform)
        self.test_transform = wrap_in_compose_is_list(test_transform)
        self.predict_transform = wrap_in_compose_is_list(predict_transform)
        self.data_root = data_root
        self.scenario = scenario
        self.requires_norm = requires_norm
        self.binary_labels = binary_labels
        self.linear_encoder = linear_encoder
        self.kwargs = kwargs

    def setup(self, stage: str) -> None:
        """Set up datasets.

        Args:
            stage: Either fit, validate, test, or predict.
        """
        if stage in ["fit"]:
            self.train_dataset = Sen4AgriNet(
                split="train",
                data_root=self.data_root,
                transform=self.train_transform,
                bands=self.bands,
                seed=self.seed,
                scenario=self.scenario,
                requires_norm=self.requires_norm,
                binary_labels=self.binary_labels,
                linear_encoder=self.linear_encoder,
                **self.kwargs,
            )
        if stage in ["fit", "validate"]:
            self.val_dataset = Sen4AgriNet(
                split="val",
                data_root=self.data_root,
                transform=self.val_transform,
                bands=self.bands,
                seed=self.seed,
                scenario=self.scenario,
                requires_norm=self.requires_norm,
                binary_labels=self.binary_labels,
                linear_encoder=self.linear_encoder,
                **self.kwargs,
            )
        if stage in ["test"]:
            self.test_dataset = Sen4AgriNet(
                split="test",
                data_root=self.data_root,
                transform=self.test_transform,
                bands=self.bands,
                seed=self.seed,
                scenario=self.scenario,
                requires_norm=self.requires_norm,
                binary_labels=self.binary_labels,
                linear_encoder=self.linear_encoder,
                **self.kwargs,
            )
        if stage in ["predict"]:
            self.predict_dataset = Sen4AgriNet(
                split="test",
                data_root=self.data_root,
                transform=self.predict_transform,
                bands=self.bands,
                seed=self.seed,
                scenario=self.scenario,
                requires_norm=self.requires_norm,
                binary_labels=self.binary_labels,
                linear_encoder=self.linear_encoder,
                **self.kwargs,
            )



# Copyright contributors to the Terratorch project

"""Module containing generic multimodal dataset classes"""

import glob
import logging
import random
import warnings
import os
import re
import torch
import rasterio
import pandas as pd
from abc import ABC
from pathlib import Path
from typing import Any

import albumentations as A
import numpy as np
import rioxarray
import xarray as xr
from einops import rearrange
from matplotlib import pyplot as plt
from matplotlib.figure import Figure
from matplotlib.patches import Rectangle
from matplotlib.colors import ListedColormap
from torchgeo.datasets import NonGeoDataset

from terratorch.datasets.utils import default_transform, generate_bands_intervals
from terratorch.datasets.transforms import MultimodalTransforms, MultimodalToTensor

logger = logging.getLogger("terratorch")


def load_table_data(file_path: str | Path) -> pd.DataFrame:
    file_path = str(file_path)
    if file_path.endswith("parquet"):
        df = pd.read_parquet(file_path)
    elif file_path.endswith("csv"):
        df = pd.read_csv(file_path, index_col=0)
    else:
        raise Exception(f"Unrecognized file type: {file_path}. Only parquet and csv are supported.")
    return df


class GenericMultimodalDataset(NonGeoDataset, ABC):
    """
    This is a generic dataset class initialized by
    [GenericMultiModalDataModule][terratorch.datamodules.GenericMultiModalDataModule].
    """

    def __init__(
        self,
        data_root: dict[str, Path | str],
        label_data_root: Path | str | list[Path | str] | None = None,
        image_grep: dict[str, str] | None = "*",
        label_grep: str | None = "*",
        prefix: str | None = "*",
        split: Path | None = None,
        image_modalities: list[str] | None = None,
        rgb_indices: dict[str, list[int]] | None = None,
        allow_missing_modalities: bool = False,
        allow_substring_file_names: bool = True,
        skip_file_checks: bool = False,
        dataset_bands: dict[str, list] | None = None,
        output_bands: dict[str, list] | None = None,
        constant_scale: dict[str, float] = None,
        transform: A.Compose | dict | None = None,
        no_data_replace: float | None = None,
        no_label_replace: float | None = -1,
        expand_temporal_dimension: bool = False,
        temporal_channel_major: bool = False,
        reduce_zero_label: bool = False,
        channel_position: int = -3,
        scalar_label: bool = False,
        data_with_sample_dim: bool = False,
        concat_bands: bool = False,
        *args,
        **kwargs,
    ) -> None:
        """Constructor

        Args:
            data_root (dict[Path]): Dictionary of paths to data root directory or csv/parquet files with image-level
                data, with modalities as keys.
            label_data_root (Path, optional): Path to data root directory with labels or csv/parquet files with
                image-level labels. Needs to be specified for supervised tasks. Set to None for prediction mode.
            image_grep (dict[str], optional): Dictionary with regular expression appended to data_root to find input
                images, with modalities as keys. Defaults to "*". Ignored when allow_substring_file_names is False.
            label_grep (str, optional): Regular expression appended to label_data_root to find labels or mask files.
                Defaults to "*". Ignored when allow_substring_file_names is False.
             prefix (str, optional): Prefix for filenames and/ or in case of data in subdirs (e.g. "part-xxx). Defaults to "*".
            split (Path, optional): Path to file containing samples prefixes to be used for this split.
                The file can be a csv/parquet file with the prefixes in the index or a txt file with new-line separated
                sample prefixes. File names must be exact matches if allow_substring_file_names is False. Otherwise,
                files are searched using glob with the form Path(data_root).glob(prefix + [image or label grep]).
                If not specified, search samples based on files in data_root. Defaults to None.
            image_modalities(list[str], optional): List of pixel-level raster modalities. Defaults to data_root.keys().
                The difference between all modalities and image_modalities are non-image modalities which are treated
                differently during the transforms and are not modified but only converted into a tensor if possible.
            rgb_indices (dict[str, list[int]], optional): Indices of RGB channels for plotting with the format
                {<modality>: [<band indices>]}. Defaults to {image_modalities[0]: [0, 1, 2]} if not provided.
            allow_missing_modalities (bool, optional): Allow missing modalities during data loading. Defaults to False.
            allow_substring_file_names (bool, optional): Allow substrings during sample identification using
                wildcards (*). If False, treats sample prefix + image_grep as full file name. Defaults to True.
            dataset_bands (dict[list], optional): Bands present in the dataset, provided in a dictionary with modalities
                as keys. This parameter names input channels (bands) using HLSBands, ints, int ranges, or strings, so
                that they can then be referred to by output_bands. Needs to be superset of output_bands. Can be a subset
                of all modalities. Defaults to None.
            output_bands (dict[list], optional): Bands that should be output by the dataset as named by dataset_bands,
                provided as a dictionary with modality keys. Can be subset of all modalities. Defaults to None.
            constant_scale (dict[str, float]): Factor to multiply data values by, provided as a dictionary with
                modalities as keys. Can be subset of all modalities. Defaults to None.
            transform (Albumentations.Compose | dict | None): Albumentations transform to be applied to all image
                modalities (transformation are shared between image modalities, e.g., similar crop or rotation).
                Should end with albumentations.ToTensorV2(). If used through the generic_data_module, should not include
                normalization. The transform is not applied to non-image data, which is only converted to tensors if
                possible. If dict with modalities as keys, transforms are applied seperatly per modality without shared
                parameters. Defaults to terratorch.datasets.transforms.MultimodalToTensor().
            no_data_replace (float | None): Replace nan values in input data with this value.
                If None, does no replacement. Defaults to None.
            no_label_replace (float | None): Replace nan values in label with this value.
                If none, does no replacement. Defaults to -1.
            skip_file_checks (bool, optional): Skips the check if a sample path exists. Only works
                with allow_missing_modalities=False and allow_substring_file_names=False. Samples are expected in the
                format <prefix><image_grep> without any wildcards (*), e.g. sample1_s2l2a.tif. Defaults to False.
            expand_temporal_dimension (bool): Go from shape (time*channels, h, w) to (channels, time, h, w).
                Only works with image modalities. Is only applied to modalities with defined dataset_bands.
                Defaults to False. Assumes bands are grouped by time (all bands of one timestep are stacked together), otherwise set temporal_channel_major=True.
            temporal_channel_major: Used for expand_temporal_dimension, set True if bands are grouped by channel (all timesteps of one band are stacked together).
            reduce_zero_label (bool): Subtract 1 from all labels. Useful when labels start from 1 instead of the
                expected 0. Defaults to False.
            channel_position (int): Position of the channel dimension in the image modalities. Defaults to -3.
            scalar_label (bool): Returns a image mask if False or otherwise the raw labels. Defaults to False.
            concat_bands (bool): Concatenate all image modalities along the band dimension into a single "image", so
                that it can be processed by single-modal models. Concatenate in the order of provided modalities.
                Works with image modalities only. Does not work with allow_missing_modalities. Defaults to False.
        """

        super().__init__()

        self.split_file = split
        self.modalities = list(data_root.keys())
        assert "mask" not in self.modalities, "Modality cannot be called 'mask'."
        self.image_modalities = image_modalities or self.modalities
        self.non_image_modalities = list(set(self.modalities) - set(image_modalities))
        self.modalities = self.image_modalities + self.non_image_modalities  # Ensure image modalities to be first

        # Order by modalities and convert path strings to lists as the code expects a list of paths per modality
        data_root = {m: data_root[m] for m in self.modalities}

        # Default behaviour (no subdirs)
        self.subdir_data = False
        self.part_dirs = None
        self.prefix = prefix if prefix is not None else "*"

        self.constant_scale = constant_scale or {}
        self.no_data_replace = no_data_replace
        self.no_label_replace = no_label_replace
        self.reduce_zero_label = reduce_zero_label
        self.expand_temporal_dimension = expand_temporal_dimension
        self.temporal_channel_major = temporal_channel_major
        self.channel_position = channel_position
        self.scalar_label = scalar_label
        self.data_with_sample_dim = data_with_sample_dim
        self.concat_bands = concat_bands
        assert not self.concat_bands or len(self.non_image_modalities) == 0, (
            f"concat_bands can only be used with image modalities, "
            f"but non-image modalities are given: {self.non_image_modalities}"
        )
        assert (
            not self.concat_bands or not allow_missing_modalities
        ), "concat_bands cannot be used with allow_missing_modalities."

        if self.expand_temporal_dimension:
            if not self.temporal_channel_major:
                warnings.warn(
                    "expand_temporal_dimension=True assumes bands are grouped by time "
                    "(all bands of one timestep are stacked together). "
                    "If instead bands are grouped by channel "
                    "(all timesteps of one band are stacked together), "
                    "set temporal_channel_major=True.")

            if dataset_bands is None:
                raise ValueError(
                    "Please provide dataset_bands when expand_temporal_dimension=True."
                )

        if scalar_label:
            self.non_image_modalities += ["label"]

        # Load samples based on split file
        if self.split_file is not None:
            if not os.path.isfile(self.split_file):
                raise FileNotFoundError(f"Split file {self.split_file} does not exist.")
            elif str(self.split_file).endswith(".txt"):
                with open(self.split_file) as f:
                    split = f.readlines()
                valid_files = [rf"{substring.strip()}" for substring in split]
            else:
                valid_files = list(load_table_data(self.split_file).index)
            if len(valid_files) == 0:
                raise ValueError(f"No sample candidates (file prefixes) found in split file {self.split_file}.")

            # Prepend prefix to all filenames, and warn if wildcards remain in it
            prefix = getattr(self, "prefix", "*")
            if prefix != "*":
                # Warn if prefix contains a wildcard pattern
                if "*" in prefix:
                    warnings.warn(
                        f"Prefix={prefix!r} contains a wildcard pattern. Prepending it to split-file entries may result in a heavy glob search later. "
                        f"If the split file entries already contain the prefix, do not pass the `prefix` argument. Otherwise, provide a non-glob prefix "
                        f"(e.g. 'part-000013/', 'files_') or consider adding prefix into split file entries.",
                        stacklevel=2,
                    )

                # Prepend prefix, make sure v is not an absolute path
                valid_files = [
                    (os.path.join(prefix, v.lstrip("/\\")) if v and not os.path.isabs(v) else v)
                    for v in valid_files
                ]

        else:
            image_files = {}
            for m, m_paths in data_root.items():
                image_files[m] = sorted(glob.glob(os.path.join(m_paths, self.prefix + image_grep[m])))
                if len(image_files[m]) > 10_000:
                    warnings.warn("Found large data folder, consider providing split files to speed up dataset build.")

            def get_file_id(file_name, mod):
                base = os.path.basename(file_name)
                glob_as_regex = '^(.*?)' + ''.join(re.escape(ch) for ch in image_grep[mod].strip('*')) + '$'
                stem = re.match(glob_as_regex, base).group(1)

                if "." not in image_grep[mod] and allow_substring_file_names:
                    stem = os.path.splitext(stem)[0]

                if "/" in self.prefix:
                    parent = os.path.basename(os.path.dirname(file_name)) # direct parent dir

                    # warn if we have more than one subdirectory level
                    if self.prefix.count("/") > 1:
                        warnings.warn(
                            f"Multiple subdirectories in {self.prefix!r}; only a single parent directory "
                            f"is supported. Using parent={parent!r}.",
                            stacklevel=2,
                        )
                    return f"{parent}/{stem}"

                return stem

            if allow_missing_modalities:
                valid_files = list(set([get_file_id(file, mod)
                                        for mod, files in image_files.items()
                                        for file in files
                                        ]))
            else:
                valid_files = [get_file_id(file, self.modalities[0]) for file in image_files[self.modalities[0]]]

        self.samples = []
        num_modalities = len(self.modalities) + int(label_data_root is not None)
        if len(valid_files) == 0:
            # Provide additional information if no candidates are found
            image_files = {m: f[:3] for m, f in image_files.items()}
            raise ValueError(f"No sample candidates (file prefixes) found for multimodal dataset. "
                             f"Please review files and parameters.\n"
                             f"data_root: {data_root}\n"
                             f"image_grep: {image_grep}\n"
                             f"allow_missing_modalities: {allow_missing_modalities}\n"
                             f"File examples in data_root: {image_files}\n")

        # Check for parquet and csv files with modality data and read the file
        for m, m_path in data_root.items():
            if os.path.isfile(m_path):
                data_root[m] = load_table_data(m_path)
                # Check for some sample keys
                if not any(f in data_root[m].index for f in valid_files[:100]):
                    warnings.warn(f"Sample key expected in table index (first column) for {m} (file: {m_path}). "
                                  f"{valid_files[:3]+['...']} are not in index {list(data_root[m].index[:3])+['...']}.")
        if label_data_root is not None:
            if os.path.isfile(label_data_root):
                label_data_root = load_table_data(label_data_root)
                # Check for some sample keys
                if not any(f in label_data_root.index for f in valid_files[:100]):
                    warnings.warn(f"Keys expected in table index (first column) for labels (file: {label_data_root}). "
                                  f"The keys {valid_files[:3] + ['...']} are not in the index.")

        # Iterate over all files in split
        failed_candidates = []
        for file in valid_files:
            sample = {}
            # Iterate over all modalities
            for m, m_path in data_root.items():
                if isinstance(m_path, pd.DataFrame):
                    # Add tabular data to sample
                    sample[m] = m_path.loc[file].values
                elif allow_substring_file_names:
                    # Substring match with image_grep
                    if os.path.exists(os.path.join(m_path, file + image_grep[m].strip('*'))):
                        # Avoid glob if possible to speed up the dataset build
                        m_files = [os.path.join(m_path, file + image_grep[m].strip('*'))]
                    else:
                        m_files = sorted(glob.glob(os.path.join(m_path, file + image_grep[m])))
                        if len(valid_files) > 10_000:
                            warnings.warn("Found large data folder. You can speed up the dataset build by "
                                          "providing split files with sample ids and suffixes without wildcards. E.g. "
                                          "sample id 'sample1' and suffix '_s2l2a.tif' for file 'sample1_s2l2a.tif'.")

                    if m_files:
                        sample[m] = m_files[-1]
                        if len(m_files) > 1:
                            warnings.warn(f"Found multiple matching files for sample {file} and grep {image_grep[m]}: "
                                          f"{m_files}. Selecting last one. "
                                          f"Consider changing data structure or parameters for unique selection.")
                else:
                    # Exact match
                    file_path = os.path.join(m_path, file + image_grep[m].strip('*'))
                    if skip_file_checks or os.path.exists(file_path):
                        sample[m] = file_path

            if label_data_root is not None:
                if isinstance(label_data_root, pd.DataFrame):
                    # Add tabular data to sample
                    sample["mask"] = label_data_root.loc[file].values
                elif allow_substring_file_names:
                    if os.path.exists(os.path.join(label_data_root, file + label_grep.strip('*'))):
                        # Avoid glob if possible to speed up the dataset build
                        l_files = [os.path.join(label_data_root, file + label_grep.strip('*'))]
                    else:
                        # Substring match with label_grep
                        l_files = sorted(glob.glob(os.path.join(label_data_root, file + label_grep)))
                    if l_files:
                        sample["mask"] = l_files[-1]
                else:
                    # Exact match
                    file_path = os.path.join(label_data_root, file + label_grep.strip('*'))
                    if skip_file_checks or os.path.exists(file_path):
                        sample["mask"] = file_path
                if "mask" not in sample:
                    # Only add sample if mask is present
                    failed_candidates.append(sample)
                    continue

            if len(sample) == num_modalities or allow_missing_modalities:
                self.samples.append(sample)
            else:
                failed_candidates.append(sample)

        if len(self.samples) == 0:
            # Provide additional information if no multi-modal samples are found
            idx = random.sample(range(len(valid_files)), min(5, len(valid_files)))
            raise ValueError(f"No samples found for multimodal dataset. Please review files, path, and grep params.\n"
                             f"data_root: {data_root}\n"
                             f"image_grep: {image_grep}\n"
                             f"allow_substring_file_names: {allow_substring_file_names}\n"
                             f"allow_missing_modalities: {allow_missing_modalities}\n"
                             f"Candidate prefixes: {', '.join([valid_files[i] for i in idx])}\n"
                             f"Sample candidate paths: {', '.join([str(failed_candidates[i]) for i in idx])}")

        self.rgb_indices = rgb_indices or {image_modalities[0]: [0, 1, 2]}

        if dataset_bands is not None:
            self.dataset_bands = {m: generate_bands_intervals(m_bands) for m, m_bands in dataset_bands.items()}
        else:
            self.dataset_bands = None
        if output_bands is not None:
            self.output_bands = {m: generate_bands_intervals(m_bands) for m, m_bands in output_bands.items()}
            for modality in self.modalities:
                if modality in self.output_bands and modality not in self.dataset_bands:
                    msg = f"If output bands are provided, dataset_bands must also be provided (modality: {modality})"
                    raise Exception(msg)  # noqa: PLE0101
        else:
            self.output_bands = {}

        self.filter_indices = {}
        if self.output_bands:
            for m in self.output_bands.keys():
                if m not in self.output_bands or self.output_bands[m] == self.dataset_bands[m]:
                    continue
                if len(set(self.output_bands[m]) & set(self.dataset_bands[m])) != len(self.output_bands[m]):
                    msg = f"Output bands must be a subset of dataset bands (Modality: {m})"
                    raise Exception(msg)

                self.filter_indices[m] = [self.dataset_bands[m].index(band) for band in self.output_bands[m]]

            if not self.channel_position:
                logger.warning(
                    "output_bands is defined but no channel_position is provided. "
                    "Channels must be in the last dimension, otherwise provide channel_position."
                )

        # If no transform is given, apply only to transform to torch tensor
        if isinstance(transform, A.Compose):
            self.transform = MultimodalTransforms(transform,
                                                  non_image_modalities=self.non_image_modalities + ['label']
                                                  if scalar_label else self.non_image_modalities)
        elif isinstance(transform, dict):
            # Modality-specific transforms
            transform = {m: transform[m] if m in transform else default_transform for m in self.modalities}
            self.transform = MultimodalTransforms(transform, shared=False)
        elif transform is None:
            self.transform = MultimodalToTensor(self.modalities)
        else:
            raise ValueError(f'Unknown transform type: {type(transform)}, expect A.Compose, dict or None.')

        # Ignore rasterio warning for not geo-referenced files
        warnings.filterwarnings("ignore", category=rasterio.errors.NotGeoreferencedWarning)
        warnings.filterwarnings("ignore", message="Dataset has no geotransform")

    def __len__(self) -> int:
        return len(self.samples)

    def __getitem__(self, index: int) -> dict[str, Any]:
        output = {}
        if isinstance(index, tuple):
            # Load only sampled modalities instead of all modalities
            # (see sample_num_modalities in GenericMultiModalDataModule for details)
            index, modalities = index
            sample = {m: self.samples[index][m] for m in modalities}
        else:
            sample = self.samples[index]

        for modality, file in sample.items():
            data = self._load_file(
                file,
                nan_replace=self.no_label_replace if modality == "mask" else self.no_data_replace,
                modality=modality,
            )

            # Expand temporal dim
            if modality in self.filter_indices and self.expand_temporal_dimension:
                if self.temporal_channel_major:
                    data = rearrange(
                        data, "(channels time) h w -> channels time h w", channels=len(self.dataset_bands[modality])
                    )
                else:
                    data = rearrange(
                        data, "(time channels) h w -> channels time h w", channels=len(self.dataset_bands[modality])
                    )


            if modality == "mask" and not self.scalar_label:
                # tasks expect image masks without channel dim
                data = data[0]

            if modality in self.image_modalities and len(data.shape) >= 3 and self.channel_position:
                # to channels last (required by albumentations)
                data = np.moveaxis(data, self.channel_position, -1)

            if modality in self.filter_indices:
                data = data[..., self.filter_indices[modality]]

            if modality in self.constant_scale:
                data = data.astype(np.float32) * self.constant_scale[modality]

            if data.dtype == np.float64:
                data = data.astype(np.float32)

            output[modality] = data

        if "mask" in output:
            if self.reduce_zero_label:
                output["mask"] -= 1
            if self.scalar_label:
                output["label"] = output.pop("mask")

        if self.transform:
            output = self.transform(output)

        if self.concat_bands:
            # Concatenate bands of all image modalities
            data = [output.pop(m) for m in self.image_modalities if m in output]
            output["image"] = torch.cat(data, dim=1 if self.data_with_sample_dim else 0)
        else:
            # Tasks expect data to be stored in "image", moving modalities to image dict
            output["image"] = {m: output.pop(m) for m in self.modalities if m in output}

        output["filename"] = self.samples[index]

        return output

    def _load_file(self, path, nan_replace: int | float | None = None, modality: str | None = None) -> xr.DataArray:
        if isinstance(path, np.ndarray):
            # data was loaded from table and is saved in memory
            data = path
        elif path.endswith(".zarr") or path.endswith(".zarr.zip"):
            data = xr.open_zarr(path, mask_and_scale=True)
            data_var = modality if modality in data.data_vars else list(data.data_vars)[0]
            data = data[data_var].to_numpy()
        elif path.endswith(".npy"):
            data = np.load(path)
        else:
            data = rioxarray.open_rasterio(path, masked=True).to_numpy()

        if nan_replace is not None:
            data = np.nan_to_num(data, nan=nan_replace)
        return data

    def plot(self, sample: dict[str, torch.Tensor], suptitle: str | None = None) -> Figure:
        """Plot a sample from the dataset.

        Args:
            sample: a sample returned by :meth:`__getitem__`
            suptitle: optional string to use as a suptitle

        Returns:
            a matplotlib Figure with the rendered sample
        """
        suptitle = suptitle or sample.get("filename", "").split("/")[-1].split(".")[0]
        images = {}
        for mod, indices in self.rgb_indices.items():
            if "image" in sample and isinstance(sample["image"], dict) and mod in sample["image"]:
                # Move modality to sample dict
                images[mod] = sample["image"][mod]
            if mod in sample.keys():
                image = sample[mod][indices]
                # Per modality processing
                if isinstance(image, torch.Tensor):
                    image = image.numpy()
                # Normalize to 0 - 1
                image = image - image.min(axis=(-1, -2), keepdims=True)
                image = image / np.quantile(image, q=0.99, axis=(-1, -2), keepdims=True)
                image = np.clip(image, 0, 1)
                # Channel last
                image = np.moveaxis(image, 0, -1)
                if image.ndim == 4:
                    warnings.warn(f"Found time series data. Plotting only supports images, selecting the first one.")
                    image = image[0]
                images[mod] = image

        if len(images) == 0:
            warnings.warn(f"No RGB modalities found ({suptitle}). Sample keys: {list(sample.keys())}, "
                          f"Dataset rgb_indices modalities: {list(self.rgb_indices.keys())}")
            raise ValueError("No RGB images found.")

        if "mask" in sample:
            mask = sample["mask"]
            if isinstance(mask, torch.Tensor):
                mask = mask.numpy()
        else:
            mask = None

        if "prediction" in sample:
            prediction = sample["prediction"]
            if isinstance(prediction, dict):
                raise ValueError("Multiple outputs not yet supported")
            if isinstance(prediction, torch.Tensor):
                prediction = prediction.numpy()
            while prediction.ndim < 2:
                prediction = np.expand_dims(prediction, -1)
        else:
            prediction = None

        # Scalar label
        if "label" in sample:
            label = sample["label"]
            if isinstance(label, torch.Tensor):
                label = label.numpy()
            while label.ndim < 2:
                label = np.expand_dims(label, -1)
        else:
            label = None

        if hasattr(self, "num_classes"):
            # Classification, Segmentation
            vmin, vmax = 0, self.num_classes - 1
            cmap = plt.get_cmap("rainbow")
            cmap = ListedColormap(cmap(np.linspace(0.10, 1.0, self.num_classes)))  # Start with blue
            class_names = self.class_names or list(range(0, self.num_classes))
            handles = [Rectangle((0, 0), 1, 1, color=cmap(i))
                       for i in range(self.num_classes)]

            if self.no_label_replace is not None:
                if mask is not None:
                    mask[mask == self.no_label_replace] = -1
                elif label is not None:
                    label[label == self.no_label_replace] = -1
                vmin = -1
                cmap = cmap(np.arange(self.num_classes))
                cmap = ListedColormap(np.vstack(([0, 0, 0, 1], cmap)), N=len(cmap) + 1)
                class_names = ["No label"] + class_names
                handles = [Rectangle((0, 0), 1, 1, color=(0, 0, 0, 1))] + handles
        else:
            # Regression
            vmax = np.max([mask.max() if mask is not None else 0,
                           prediction.max() if prediction is not None else 0,
                           label.max() if label is not None else 0])
            vmin = np.min([mask.min() if mask is not None else 0,
                           prediction.min() if prediction is not None else 0,
                           label.min() if label is not None else 0])
            cmap = "viridis"
            class_names = handles = None

        # Plot images
        num_images = len(images) + int(mask is not None or label is not None) + int(prediction is not None)
        fig, ax = plt.subplots(1, num_images, figsize=(5*num_images, 5))

        for i, (mod, image) in enumerate(images.items()):
            ax[i].imshow(image)
            ax[i].axis("off")
            ax[i].set_title(mod)

        image = list(images.values())[0]  # First RGB modality as base image for mask

        mask_i = -1 if prediction is None else -2
        if mask is not None:
            ax[mask_i].imshow(image)
            ax[mask_i].imshow(mask, alpha=0.7, vmin=vmin, vmax=vmax, cmap=cmap, interpolation="nearest")
            ax[mask_i].axis("off")
            ax[mask_i].set_title("GT Mask")

            if class_names is not None:
                # Segmentation task
                ax[-1].legend(handles, class_names, loc="upper left", bbox_to_anchor=(1, 1))

        if prediction is not None:
            ax[-1].imshow(image)
            ax[-1].imshow(prediction, alpha=0.7, vmin=vmin, vmax=vmax, cmap=cmap, interpolation="nearest")
            ax[-1].axis("off")
            ax[-1].set_title("Prediction")

        if label is not None:
            # Plot scalar values
            ax[mask_i].imshow(image)
            ax[mask_i].imshow(label, alpha=0.7, vmin=vmin, vmax=vmax, cmap=cmap, interpolation="nearest")
            ax[mask_i].axis("off")
            ax[mask_i].set_title(f"GT Label:\n{label[:, 0]}")
            if prediction is not None:
                ax[-1].set_title(f"Prediction:\n{prediction[:, 0]}")

        if suptitle is not None:
            plt.suptitle(suptitle)
        return fig


class GenericMultimodalSegmentationDataset(GenericMultimodalDataset):
    """GenericNonGeoSegmentationDataset"""

    def __init__(
        self,
        data_root: Path,
        num_classes: int,
        label_data_root: Path | None = None,
        image_grep: str | None = "*",
        label_grep: str | None = "*",
        split: Path | None = None,
        image_modalities: list[str] | None = None,
        rgb_indices: list[str] | None = None,
        allow_missing_modalities: bool = False,
        allow_substring_file_names: bool = False,
        skip_file_checks: bool = False,
        dataset_bands: dict[str, list] | None = None,
        output_bands: dict[str, list] | None = None,
        class_names: list[str] | None = None,
        constant_scale: dict[str, float] = None,
        transform: A.Compose | None = None,
        no_data_replace: float | None = None,
        no_label_replace: int | None = -1,
        expand_temporal_dimension: bool = False,
        reduce_zero_label: bool = False,
        channel_position: int = -3,
        concat_bands: bool = False,
        *args,
        **kwargs,
    ) -> None:
        """Constructor

        Args:
            data_root (dict[Path]): Dictionary of paths to data root directory or csv/parquet files with image-level
                data, with modalities as keys.
            num_classes (int): Number of classes.
            label_data_root (Path): Path to data root directory with mask files. Set to None for prediction mode.
            image_grep (dict[str], optional): Dictionary with regular expression appended to data_root to find input
                images, with modalities as keys. Defaults to "*". Ignored when allow_substring_file_names is False.
            label_grep (str, optional): Regular expression appended to label_data_root to find mask files.
                Defaults to "*". Ignored when allow_substring_file_names is False.
            split (Path, optional): Path to file containing samples prefixes to be used for this split.
                The file can be a csv/parquet file with the prefixes in the index or a txt file with new-line separated
                sample prefixes. File names must be exact matches if allow_substring_file_names is False. Otherwise,
                files are searched using glob with the form Path(data_root).glob(prefix + [image or label grep]).
                If not specified, search samples based on files in data_root. Defaults to None.
            image_modalities(list[str], optional): List of pixel-level raster modalities. Defaults to data_root.keys().
                The difference between all modalities and image_modalities are non-image modalities which are treated
                differently during the transforms and are not modified but only converted into a tensor if possible.
            rgb_indices (dict[str, list[int]], optional): Indices of RGB channels for plotting with the format
                {<modality>: [<band indices>]}. Defaults to {image_modalities[0]: [0, 1, 2]} if not provided.
            allow_missing_modalities (bool, optional): Allow missing modalities during data loading. Defaults to False.
            allow_substring_file_names (bool, optional): Allow substrings during sample identification using
                wildcards (*). If False, treats sample prefix + image_grep as full file name. Defaults to True.
            skip_file_checks (bool, optional): Skips the check if a sample path exists. Only works
                with allow_missing_modalities=False and allow_substring_file_names=False. Samples are expected in the
                format <prefix><image_grep> without any wildcards (*), e.g. sample1_s2l2a.tif. Defaults to False.
            dataset_bands (dict[list], optional): Bands present in the dataset, provided in a dictionary with modalities
                as keys. This parameter names input channels (bands) using HLSBands, ints, int ranges, or strings, so
                that they can then be referred to by output_bands. Needs to be superset of output_bands. Can be a subset
                of all modalities. Defaults to None.
            output_bands (dict[list], optional): Bands that should be output by the dataset as named by dataset_bands,
                provided as a dictionary with modality keys. Can be subset of all modalities. Defaults to None.
            class_names (list[str], optional): Names of the classes. Defaults to None.
            constant_scale (dict[str, float]): Factor to multiply data values by, provided as a dictionary with
                modalities as keys. Can be subset of all modalities. Defaults to None.
            transform (Albumentations.Compose | dict | None): Albumentations transform to be applied to all image
                modalities (transformation are shared between image modalities, e.g., similar crop or rotation).
                Should end with albumentations.ToTensorV2(). If used through the generic_data_module, should not include
                normalization. The transform is not applied to non-image data, which is only converted to tensors if
                possible. If dict with modalities as keys, transforms are applied seperatly per modality without shared
                parameters. Defaults to terratorch.datasets.transforms.MultimodalToTensor().
            no_data_replace (float | None): Replace nan values in input data with this value.
                If None, does no replacement. Defaults to None.
            no_label_replace (float | None): Replace nan values in label with this value.
                If none, does no replacement. Defaults to -1.
            expand_temporal_dimension (bool): Go from shape (time*channels, h, w) to (channels, time, h, w).
                Only works with image modalities. Is only applied to modalities with defined dataset_bands.
                Defaults to False.
            reduce_zero_label (bool): Subtract 1 from all labels. Useful when labels start from 1 instead of the
                expected 0. Defaults to False.
            channel_position (int): Position of the channel dimension in the image modalities. Defaults to -3.
            concat_bands (bool): Concatenate all image modalities along the band dimension into a single "image", so
                that it can be processed by single-modal models. Concatenate in the order of provided modalities.
                Works with image modalities only. Does not work with allow_missing_modalities. Defaults to False.
        """

        super().__init__(
            data_root,
            label_data_root=label_data_root,
            image_grep=image_grep,
            label_grep=label_grep,
            split=split,
            image_modalities=image_modalities,
            rgb_indices=rgb_indices,
            allow_missing_modalities=allow_missing_modalities,
            allow_substring_file_names=allow_substring_file_names,
            skip_file_checks=skip_file_checks,
            dataset_bands=dataset_bands,
            output_bands=output_bands,
            constant_scale=constant_scale,
            transform=transform,
            no_data_replace=no_data_replace,
            no_label_replace=no_label_replace,
            expand_temporal_dimension=expand_temporal_dimension,
            reduce_zero_label=reduce_zero_label,
            channel_position=channel_position,
            concat_bands=concat_bands,
            *args,
            **kwargs,
        )
        self.num_classes = num_classes
        self.class_names = class_names

    def __getitem__(self, index: int) -> dict[str, Any]:
        item = super().__getitem__(index)

        if "mask" in item:
            item["mask"] = item["mask"].long()

        return item


class GenericMultimodalPixelwiseRegressionDataset(GenericMultimodalDataset):
    """GenericNonGeoPixelwiseRegressionDataset"""

    def __init__(
        self,
        data_root: Path,
        label_data_root: Path | None = None,
        image_grep: str | None = "*",
        label_grep: str | None = "*",
        split: Path | None = None,
        image_modalities: list[str] | None = None,
        rgb_indices: list[int] | None = None,
        allow_missing_modalities: bool = False,
        allow_substring_file_names: bool = False,
        skip_file_checks: bool = False,
        dataset_bands: dict[str, list] | None = None,
        output_bands: dict[str, list] | None = None,
        constant_scale: dict[str, float] = None,
        transform: A.Compose | dict | None = None,
        no_data_replace: float | None = None,
        no_label_replace: float | None = None,
        expand_temporal_dimension: bool = False,
        reduce_zero_label: bool = False,
        channel_position: int = -3,
        concat_bands: bool = False,
        *args,
        **kwargs,
    ) -> None:
        """Constructor

        Args:
            data_root (dict[Path]): Dictionary of paths to data root directory or csv/parquet files with image-level
                data, with modalities as keys.
            label_data_root (Path): Path to data root directory with ground truth files. Set to None for predictions.
            image_grep (dict[str], optional): Dictionary with regular expression appended to data_root to find input
                images, with modalities as keys. Defaults to "*". Ignored when allow_substring_file_names is False.
            label_grep (str, optional): Regular expression appended to label_data_root to find ground truth files.
                Defaults to "*". Ignored when allow_substring_file_names is False.
            split (Path, optional): Path to file containing samples prefixes to be used for this split.
                The file can be a csv/parquet file with the prefixes in the index or a txt file with new-line separated
                sample prefixes. File names must be exact matches if allow_substring_file_names is False. Otherwise,
                files are searched using glob with the form Path(data_root).glob(prefix + [image or label grep]).
                If not specified, search samples based on files in data_root. Defaults to None.
            image_modalities(list[str], optional): List of pixel-level raster modalities. Defaults to data_root.keys().
                The difference between all modalities and image_modalities are non-image modalities which are treated
                differently during the transforms and are not modified but only converted into a tensor if possible.
            rgb_indices (dict[str, list[int]], optional): Indices of RGB channels for plotting with the format
                {<modality>: [<band indices>]}. Defaults to {image_modalities[0]: [0, 1, 2]} if not provided.
            allow_missing_modalities (bool, optional): Allow missing modalities during data loading. Defaults to False.
            allow_substring_file_names (bool, optional): Allow substrings during sample identification using
                wildcards (*). If False, treats sample prefix + image_grep as full file name. Defaults to True.
            skip_file_checks (bool, optional): Skips the check if a sample path exists. Only works
                with allow_missing_modalities=False and allow_substring_file_names=False. Samples are expected in the
                format <prefix><image_grep> without any wildcards (*), e.g. sample1_s2l2a.tif. Defaults to False.
            dataset_bands (dict[list], optional): Bands present in the dataset, provided in a dictionary with modalities
                as keys. This parameter names input channels (bands) using HLSBands, ints, int ranges, or strings, so
                that they can then be referred to by output_bands. Needs to be superset of output_bands. Can be a subset
                of all modalities. Defaults to None.
            output_bands (dict[list], optional): Bands that should be output by the dataset as named by dataset_bands,
                provided as a dictionary with modality keys. Can be subset of all modalities. Defaults to None.
            constant_scale (dict[float]): Factor to multiply data values by, provided as a dictionary with modalities as
                keys. Can be subset of all modalities. Defaults to None.
            transform (Albumentations.Compose | dict | None): Albumentations transform to be applied to all image
                modalities (transformation are shared between image modalities, e.g., similar crop or rotation).
                Should end with albumentations.ToTensorV2(). If used through the generic_data_module, should not include
                normalization. The transform is not applied to non-image data, which is only converted to tensors if
                possible. If dict with modalities as keys, transforms are applied seperatly per modality without shared
                parameters. Defaults to terratorch.datasets.transforms.MultimodalToTensor().
            no_data_replace (float | None): Replace nan values in input data with this value.
                If None, does no replacement. Defaults to None.
            no_label_replace (float | None): Replace nan values in label with this value.
                If none, does no replacement. Defaults to None.
            expand_temporal_dimension (bool): Go from shape (time*channels, h, w) to (channels, time, h, w).
                Only works with image modalities. Is only applied to modalities with defined dataset_bands.
                Defaults to False.
            reduce_zero_label (bool): Subtract 1 from all labels. Useful when labels start from 1 instead of the
                expected 0. Defaults to False.
            channel_position (int): Position of the channel dimension in the image modalities. Defaults to -3.
            concat_bands (bool): Concatenate all image modalities along the band dimension into a single "image", so
                that it can be processed by single-modal models. Concatenate in the order of provided modalities.
                Works with image modalities only. Does not work with allow_missing_modalities. Defaults to False.
        """

        super().__init__(
            data_root,
            label_data_root=label_data_root,
            image_grep=image_grep,
            label_grep=label_grep,
            split=split,
            image_modalities=image_modalities,
            rgb_indices=rgb_indices,
            allow_missing_modalities=allow_missing_modalities,
            allow_substring_file_names=allow_substring_file_names,
            skip_file_checks=skip_file_checks,
            dataset_bands=dataset_bands,
            output_bands=output_bands,
            constant_scale=constant_scale,
            transform=transform,
            no_data_replace=no_data_replace,
            no_label_replace=no_label_replace,
            expand_temporal_dimension=expand_temporal_dimension,
            reduce_zero_label=reduce_zero_label,
            channel_position=channel_position,
            concat_bands=concat_bands,
            *args,
            **kwargs,
        )

    def __getitem__(self, index: int) -> dict[str, Any]:
        item = super().__getitem__(index)

        if "mask" in item:
            item["mask"] = item["mask"].float()

        return item


class GenericMultimodalScalarDataset(GenericMultimodalDataset):
    """GenericMultimodalClassificationDataset"""

    def __init__(
        self,
        data_root: Path,
        num_classes: int,
        label_data_root: Path | None = None,
        image_grep: str | None = "*",
        label_grep: str | None = "*",
        split: Path | None = None,
        image_modalities: list[str] | None = None,
        rgb_indices: list[int] | None = None,
        allow_missing_modalities: bool = False,
        allow_substring_file_names: bool = False,
        skip_file_checks: bool = False,
        dataset_bands: dict[str, list] | None = None,
        output_bands: dict[str, list] | None = None,
        class_names: list[str] | None = None,
        constant_scale: dict[str, float] = None,
        transform: A.Compose | None = None,
        no_data_replace: float | None = None,
        no_label_replace: int | None = None,
        expand_temporal_dimension: bool = False,
        reduce_zero_label: bool = False,
        channel_position: int = -3,
        concat_bands: bool = False,
        *args,
        **kwargs,
    ) -> None:
        """Constructor

        Args:
            data_root (dict[Path]): Dictionary of paths to data root directory or csv/parquet files with image-level
                data, with modalities as keys.
            num_classes (int): Number of classes.
            label_data_root (Path, optional): Path to data root directory with labels or csv/parquet files with labels.
                Set to None for prediction mode.
            image_grep (dict[str], optional): Dictionary with regular expression appended to data_root to find input
                images, with modalities as keys. Defaults to "*". Ignored when allow_substring_file_names is False.
            label_grep (str, optional): Regular expression appended to label_data_root to find labels files.
                Defaults to "*". Ignored when allow_substring_file_names is False.
            split (Path, optional): Path to file containing samples prefixes to be used for this split.
                The file can be a csv/parquet file with the prefixes in the index or a txt file with new-line separated
                sample prefixes. File names must be exact matches if allow_substring_file_names is False. Otherwise,
                files are searched using glob with the form Path(data_root).glob(prefix + [image or label grep]).
                If not specified, search samples based on files in data_root. Defaults to None.
            image_modalities(list[str], optional): List of pixel-level raster modalities. Defaults to data_root.keys().
                The difference between all modalities and image_modalities are non-image modalities which are treated
                differently during the transforms and are not modified but only converted into a tensor if possible.
            rgb_indices (dict[str, list[int]], optional): Indices of RGB channels for plotting with the format
                {<modality>: [<band indices>]}. Defaults to {image_modalities[0]: [0, 1, 2]} if not provided.
            allow_missing_modalities (bool, optional): Allow missing modalities during data loading. Defaults to False.
            allow_substring_file_names (bool, optional): Allow substrings during sample identification using
                wildcards (*). If False, treats sample prefix + image_grep as full file name. Defaults to True.
            skip_file_checks (bool, optional): Skips the check if a sample path exists. Only works
                with allow_missing_modalities=False and allow_substring_file_names=False. Samples are expected in the
                format <prefix><image_grep> without any wildcards (*), e.g. sample1_s2l2a.tif. Defaults to False.
            dataset_bands (dict[list], optional): Bands present in the dataset, provided in a dictionary with modalities
                as keys. This parameter names input channels (bands) using HLSBands, ints, int ranges, or strings, so
                that they can then be referred to by output_bands. Needs to be superset of output_bands. Can be a subset
                of all modalities. Defaults to None.
            output_bands (dict[list], optional): Bands that should be output by the dataset as named by dataset_bands,
                provided as a dictionary with modality keys. Can be subset of all modalities. Defaults to None.
            class_names (list[str], optional): Names of the classes. Defaults to None.
            constant_scale (dict[str, float]): Factor to multiply data values by, provided as a dictionary with
                modalities as keys. Can be subset of all modalities. Defaults to None.
            transform (Albumentations.Compose | dict | None): Albumentations transform to be applied to all image
                modalities (transformation are shared between image modalities, e.g., similar crop or rotation).
                Should end with albumentations.ToTensorV2(). If used through the generic_data_module, should not include
                normalization. The transform is not applied to non-image data, which is only converted to tensors if
                possible. If dict with modalities as keys, transforms are applied seperatly per modality without shared
                parameters. Defaults to terratorch.datasets.transforms.MultimodalToTensor().
            no_data_replace (float | None): Replace nan values in input data with this value.
                If None, does no replacement. Defaults to None.
            no_label_replace (float | None): Replace nan values in label with this value.
                If none, does no replacement. Defaults to -1.
            expand_temporal_dimension (bool): Go from shape (time*channels, h, w) to (channels, time, h, w).
                Only works with image modalities. Is only applied to modalities with defined dataset_bands.
                Defaults to False.
            reduce_zero_label (bool): Subtract 1 from all labels. Useful when labels start from 1 instead of the
                expected 0. Defaults to False.
            channel_position (int): Position of the channel dimension in the image modalities. Defaults to -3.
            concat_bands (bool): Concatenate all image modalities along the band dimension into a single "image", so
                that it can be processed by single-modal models. Concatenate in the order of provided modalities.
                Works with image modalities only. Does not work with allow_missing_modalities. Defaults to False.
        """

        super().__init__(
            data_root,
            label_data_root=label_data_root,
            image_grep=image_grep,
            label_grep=label_grep,
            split=split,
            image_modalities=image_modalities,
            rgb_indices=rgb_indices,
            allow_missing_modalities=allow_missing_modalities,
            allow_substring_file_names=allow_substring_file_names,
            skip_file_checks=skip_file_checks,
            dataset_bands=dataset_bands,
            output_bands=output_bands,
            constant_scale=constant_scale,
            transform=transform,
            no_data_replace=no_data_replace,
            no_label_replace=no_label_replace,
            expand_temporal_dimension=expand_temporal_dimension,
            reduce_zero_label=reduce_zero_label,
            channel_position=channel_position,
            scalar_label=True,
            concat_bands=concat_bands,
            *args,
            **kwargs,
        )

        self.num_classes = num_classes
        self.class_names = class_names

    def __getitem__(self, index: int) -> dict[str, Any]:
        item = super().__getitem__(index)
        return item


"""Comprehensive tests for ElephantCocoDataset.

Tests cover initialization, getitem, bbox handling, edge cases to ensure maximum code coverage.
"""

import gc
import json
import tempfile
from pathlib import Path
from unittest.mock import Mock, patch

import pytest
import torch
from PIL import Image
from torchvision import transforms as T

from terratorch.datasets.od_aed_elephant import ElephantCocoDataset


@pytest.fixture
def temp_coco_dataset():
    """Create a temporary COCO dataset structure with images and annotations."""
    with tempfile.TemporaryDirectory() as tmpdir:
        tmpdir = Path(tmpdir)
        
        # Create images directory
        img_dir = tmpdir / "images"
        img_dir.mkdir()
        
        # Create sample images
        for i in range(5):
            img = Image.new('RGB', (640, 480), color=(i*50, 100, 150))
            img.save(img_dir / f"image_{i}.jpg")
        
        # Create COCO annotations
        annotations = {
            "images": [
                {"id": 0, "file_name": "image_0.jpg", "width": 640, "height": 480},
                {"id": 1, "file_name": "image_1.jpg", "width": 640, "height": 480},
                {"id": 2, "file_name": "image_2.jpg", "width": 640, "height": 480},
                {"id": 3, "file_name": "image_3.jpg", "width": 640, "height": 480},
                {"id": 4, "file_name": "image_4.jpg", "width": 640, "height": 480},
            ],
            "annotations": [
                # Image 0: normal box
                {"id": 0, "image_id": 0, "category_id": 1, "bbox": [100, 100, 200, 150], "area": 30000, "iscrowd": 0},
                # Image 1: multiple boxes
                {"id": 1, "image_id": 1, "category_id": 1, "bbox": [50, 50, 100, 100], "area": 10000, "iscrowd": 0},
                {"id": 2, "image_id": 1, "category_id": 2, "bbox": [300, 200, 150, 100], "area": 15000, "iscrowd": 0},
                # Image 2: zero width box (should be filtered)
                {"id": 3, "image_id": 2, "category_id": 1, "bbox": [100, 100, 0, 150], "area": 0, "iscrowd": 0},
                # Image 3: zero height box (should be filtered)
                {"id": 4, "image_id": 3, "category_id": 1, "bbox": [100, 100, 200, 0], "area": 0, "iscrowd": 0},
                # Image 4: negative dimensions (should be filtered)
                {"id": 5, "image_id": 4, "category_id": 1, "bbox": [100, 100, -50, 150], "area": 0, "iscrowd": 0},
            ],
            "categories": [
                {"id": 1, "name": "elephant"},
                {"id": 2, "name": "other"},
            ]
        }
        
        ann_file = tmpdir / "annotations.json"
        with open(ann_file, 'w') as f:
            json.dump(annotations, f)
        
        yield str(img_dir), str(ann_file)


@pytest.fixture
def temp_empty_annotations():
    """Create a dataset with images but empty annotations."""
    with tempfile.TemporaryDirectory() as tmpdir:
        tmpdir = Path(tmpdir)
        
        img_dir = tmpdir / "images"
        img_dir.mkdir()
        
        # Create sample image
        img = Image.new('RGB', (640, 480), color=(100, 100, 100))
        img.save(img_dir / "image_0.jpg")
        
        # Empty annotations
        annotations = {
            "images": [
                {"id": 0, "file_name": "image_0.jpg", "width": 640, "height": 480},
            ],
            "annotations": [],
            "categories": [
                {"id": 1, "name": "elephant"},
            ]
        }
        
        ann_file = tmpdir / "annotations.json"
        with open(ann_file, 'w') as f:
            json.dump(annotations, f)
        
        yield str(img_dir), str(ann_file)


class TestElephantCocoDatasetInitialization:
    """Test suite for ElephantCocoDataset initialization."""

    def test_basic_initialization(self, temp_coco_dataset):
        """Test basic initialization without transform."""
        img_folder, ann_file = temp_coco_dataset
        
        dataset = ElephantCocoDataset(img_folder, ann_file)
        
        assert dataset is not None
        assert len(dataset) == 5
        assert dataset.transform is None

    def test_initialization_with_transform(self, temp_coco_dataset):
        """Test initialization with custom transform."""
        img_folder, ann_file = temp_coco_dataset
        transform = T.Compose([
            T.Resize((224, 224)),
            T.ColorJitter(brightness=0.2)
        ])
        
        dataset = ElephantCocoDataset(img_folder, ann_file, transform=transform)
        
        assert dataset.transform is not None
        assert len(dataset) == 5

    def test_initialization_with_none_transform(self, temp_coco_dataset):
        """Test initialization with explicit None transform."""
        img_folder, ann_file = temp_coco_dataset
        
        dataset = ElephantCocoDataset(img_folder, ann_file, transform=None)
        
        assert dataset.transform is None


class TestElephantCocoDatasetGetItem:
    """Test suite for __getitem__ method."""

    def test_getitem_single_box(self, temp_coco_dataset):
        """Test getting item with single bounding box."""
        img_folder, ann_file = temp_coco_dataset
        dataset = ElephantCocoDataset(img_folder, ann_file)
        
        sample = dataset[0]
        
        assert "image" in sample
        assert "boxes" in sample
        assert "labels" in sample
        assert "image_id" in sample
        
        assert isinstance(sample["image"], torch.Tensor)
        assert isinstance(sample["boxes"], torch.Tensor)
        assert isinstance(sample["labels"], torch.Tensor)
        assert isinstance(sample["image_id"], torch.Tensor)
        
        # Check dimensions
        assert sample["image"].shape[0] == 3  # RGB channels
        assert sample["boxes"].shape == torch.Size([1, 4])
        assert sample["labels"].shape == torch.Size([1])

    def test_getitem_multiple_boxes(self, temp_coco_dataset):
        """Test getting item with multiple bounding boxes."""
        img_folder, ann_file = temp_coco_dataset
        dataset = ElephantCocoDataset(img_folder, ann_file)
        
        sample = dataset[1]
        
        # Image 1 has 2 valid boxes
        assert sample["boxes"].shape[0] == 2
        assert sample["labels"].shape[0] == 2
        
        # Check box format (x1, y1, x2, y2)
        boxes = sample["boxes"]
        for box in boxes:
            assert box[2] > box[0]  # x2 > x1
            assert box[3] > box[1]  # y2 > y1

    def test_getitem_zero_width_box(self, temp_coco_dataset):
        """Test that boxes with zero width are filtered out."""
        img_folder, ann_file = temp_coco_dataset
        dataset = ElephantCocoDataset(img_folder, ann_file)
        
        sample = dataset[2]  # Image 2 has zero-width box
        
        # Should have no boxes
        assert sample["boxes"].shape == torch.Size([0, 4])
        assert sample["labels"].shape == torch.Size([0])

    def test_getitem_zero_height_box(self, temp_coco_dataset):
        """Test that boxes with zero height are filtered out."""
        img_folder, ann_file = temp_coco_dataset
        dataset = ElephantCocoDataset(img_folder, ann_file)
        
        sample = dataset[3]  # Image 3 has zero-height box
        
        # Should have no boxes
        assert sample["boxes"].shape == torch.Size([0, 4])
        assert sample["labels"].shape == torch.Size([0])

    def test_getitem_negative_dimensions(self, temp_coco_dataset):
        """Test that boxes with negative dimensions are filtered out."""
        img_folder, ann_file = temp_coco_dataset
        dataset = ElephantCocoDataset(img_folder, ann_file)
        
        sample = dataset[4]  # Image 4 has negative width
        
        # Should have no boxes
        assert sample["boxes"].shape == torch.Size([0, 4])
        assert sample["labels"].shape == torch.Size([0])

    def test_getitem_empty_annotations(self, temp_empty_annotations):
        """Test getting item when no annotations exist."""
        img_folder, ann_file = temp_empty_annotations
        dataset = ElephantCocoDataset(img_folder, ann_file)
        
        sample = dataset[0]
        
        # Should have empty tensors
        assert sample["boxes"].shape == torch.Size([0, 4])
        assert sample["labels"].shape == torch.Size([0])
        assert sample["boxes"].dtype == torch.float32
        assert sample["labels"].dtype == torch.int64

    def test_getitem_with_transform(self, temp_coco_dataset):
        """Test that transform is applied to image."""
        img_folder, ann_file = temp_coco_dataset
        transform = T.Resize((224, 224))
        dataset = ElephantCocoDataset(img_folder, ann_file, transform=transform)
        
        sample = dataset[0]
        
        # After ToTensor, shape is (C, H, W)
        # Transform should have resized it
        assert sample["image"].shape[1] == 224
        assert sample["image"].shape[2] == 224

    def test_getitem_image_format(self, temp_coco_dataset):
        """Test that image is correctly converted to tensor."""
        img_folder, ann_file = temp_coco_dataset
        dataset = ElephantCocoDataset(img_folder, ann_file)
        
        sample = dataset[0]
        
        # Check tensor properties
        assert sample["image"].dtype == torch.float32
        assert sample["image"].min() >= 0.0
        assert sample["image"].max() <= 1.0
        assert sample["image"].shape[0] == 3  # RGB

    def test_getitem_box_coordinates(self, temp_coco_dataset):
        """Test that box coordinates are correctly converted from COCO format."""
        img_folder, ann_file = temp_coco_dataset
        dataset = ElephantCocoDataset(img_folder, ann_file)
        
        sample = dataset[0]
        
        # COCO format: [100, 100, 200, 150] (x, y, w, h)
        # Expected: [100, 100, 300, 250] (x1, y1, x2, y2)
        box = sample["boxes"][0]
        
        assert box[0] == 100  # x1
        assert box[1] == 100  # y1
        assert box[2] == 300  # x2 = x1 + w
        assert box[3] == 250  # y2 = y1 + h

    def test_getitem_labels_format(self, temp_coco_dataset):
        """Test that labels are correctly extracted."""
        img_folder, ann_file = temp_coco_dataset
        dataset = ElephantCocoDataset(img_folder, ann_file)
        
        sample = dataset[1]  # Image with 2 boxes
        
        # Check labels
        assert sample["labels"][0] == 1  # First category
        assert sample["labels"][1] == 2  # Second category
        assert sample["labels"].dtype == torch.int64

    def test_getitem_image_id(self, temp_coco_dataset):
        """Test that image_id is correctly set."""
        img_folder, ann_file = temp_coco_dataset
        dataset = ElephantCocoDataset(img_folder, ann_file)
        
        for i in range(len(dataset)):
            sample = dataset[i]
            # image_id should match the index
            assert sample["image_id"].item() == i


class TestElephantCocoDatasetEdgeCases:
    """Test edge cases and special scenarios."""

    def test_multiple_access_same_item(self, temp_coco_dataset):
        """Test accessing the same item multiple times."""
        img_folder, ann_file = temp_coco_dataset
        dataset = ElephantCocoDataset(img_folder, ann_file)
        
        sample1 = dataset[0]
        sample2 = dataset[0]
        
        # Should return identical data
        assert torch.equal(sample1["boxes"], sample2["boxes"])
        assert torch.equal(sample1["labels"], sample2["labels"])

    def test_all_valid_indices(self, temp_coco_dataset):
        """Test that all indices can be accessed."""
        img_folder, ann_file = temp_coco_dataset
        dataset = ElephantCocoDataset(img_folder, ann_file)
        
        for i in range(len(dataset)):
            sample = dataset[i]
            assert "image" in sample
            assert "boxes" in sample
            assert "labels" in sample

    def test_box_tensor_dtype(self, temp_coco_dataset):
        """Test that boxes have correct dtype."""
        img_folder, ann_file = temp_coco_dataset
        dataset = ElephantCocoDataset(img_folder, ann_file)
        
        sample = dataset[0]
        
        assert sample["boxes"].dtype == torch.float32

    def test_label_tensor_dtype(self, temp_coco_dataset):
        """Test that labels have correct dtype."""
        img_folder, ann_file = temp_coco_dataset
        dataset = ElephantCocoDataset(img_folder, ann_file)
        
        sample = dataset[0]
        
        assert sample["labels"].dtype == torch.int64

    def test_empty_boxes_shape(self, temp_coco_dataset):
        """Test that empty boxes have correct shape."""
        img_folder, ann_file = temp_coco_dataset
        dataset = ElephantCocoDataset(img_folder, ann_file)
        
        sample = dataset[2]  # Has no valid boxes
        
        # Should be (0, 4) not (0,) or something else
        assert sample["boxes"].shape == torch.Size([0, 4])
        assert sample["boxes"].dim() == 2

    def test_empty_labels_shape(self, temp_coco_dataset):
        """Test that empty labels have correct shape."""
        img_folder, ann_file = temp_coco_dataset
        dataset = ElephantCocoDataset(img_folder, ann_file)
        
        sample = dataset[2]  # Has no valid boxes
        
        # Should be (0,)
        assert sample["labels"].shape == torch.Size([0])
        assert sample["labels"].dim() == 1

    def test_box_validity_constraints(self, temp_coco_dataset):
        """Test that all returned boxes satisfy validity constraints."""
        img_folder, ann_file = temp_coco_dataset
        dataset = ElephantCocoDataset(img_folder, ann_file)
        
        for i in range(len(dataset)):
            sample = dataset[i]
            boxes = sample["boxes"]
            
            if len(boxes) > 0:
                # All boxes should have positive width and height
                widths = boxes[:, 2] - boxes[:, 0]
                heights = boxes[:, 3] - boxes[:, 1]
                
                assert torch.all(widths > 0)
                assert torch.all(heights > 0)


class TestElephantCocoDatasetTransforms:
    """Test suite for various transforms."""

    def test_resize_transform(self, temp_coco_dataset):
        """Test with resize transform."""
        img_folder, ann_file = temp_coco_dataset
        transform = T.Resize((300, 400))
        dataset = ElephantCocoDataset(img_folder, ann_file, transform=transform)
        
        sample = dataset[0]
        
        assert sample["image"].shape[1] == 300
        assert sample["image"].shape[2] == 400

    def test_color_jitter_transform(self, temp_coco_dataset):
        """Test with color jitter transform."""
        img_folder, ann_file = temp_coco_dataset
        transform = T.ColorJitter(brightness=0.5, contrast=0.5)
        dataset = ElephantCocoDataset(img_folder, ann_file, transform=transform)
        
        sample = dataset[0]
        
        # Should still have valid image
        assert sample["image"].shape[0] == 3
        assert torch.isfinite(sample["image"]).all()

    def test_compose_transforms(self, temp_coco_dataset):
        """Test with composed transforms."""
        img_folder, ann_file = temp_coco_dataset
        transform = T.Compose([
            T.Resize((256, 256)),
            T.CenterCrop((224, 224)),
            T.ColorJitter(brightness=0.2)
        ])
        dataset = ElephantCocoDataset(img_folder, ann_file, transform=transform)
        
        sample = dataset[0]
        
        assert sample["image"].shape[1] == 224
        assert sample["image"].shape[2] == 224

    def test_grayscale_transform(self, temp_coco_dataset):
        """Test with grayscale transform."""
        img_folder, ann_file = temp_coco_dataset
        transform = T.Grayscale(num_output_channels=3)
        dataset = ElephantCocoDataset(img_folder, ann_file, transform=transform)
        
        sample = dataset[0]
        
        # Should still have 3 channels after conversion
        assert sample["image"].shape[0] == 3


class TestElephantCocoDatasetIntegration:
    """Integration tests for ElephantCocoDataset."""

    def test_dataset_length(self, temp_coco_dataset):
        """Test that dataset length matches number of images."""
        img_folder, ann_file = temp_coco_dataset
        dataset = ElephantCocoDataset(img_folder, ann_file)
        
        assert len(dataset) == 5

    def test_iteration_over_dataset(self, temp_coco_dataset):
        """Test iterating over entire dataset."""
        img_folder, ann_file = temp_coco_dataset
        dataset = ElephantCocoDataset(img_folder, ann_file)
        
        count = 0
        for sample in dataset:
            assert "image" in sample
            assert "boxes" in sample
            assert "labels" in sample
            assert "image_id" in sample
            count += 1
        
        assert count == len(dataset)

    def test_dataloader_compatibility(self, temp_coco_dataset):
        """Test that dataset works with DataLoader."""
        from torch.utils.data import DataLoader
        
        img_folder, ann_file = temp_coco_dataset
        dataset = ElephantCocoDataset(img_folder, ann_file)
        
        # Custom collate function for variable-sized boxes
        def collate_fn(batch):
            return batch
        
        dataloader = DataLoader(
            dataset, 
            batch_size=2, 
            shuffle=False,
            collate_fn=collate_fn
        )
        
        batch = next(iter(dataloader))
        assert len(batch) == 2
        
        gc.collect()

    def test_consistency_across_epochs(self, temp_coco_dataset):
        """Test that data is consistent across multiple epochs."""
        img_folder, ann_file = temp_coco_dataset
        dataset = ElephantCocoDataset(img_folder, ann_file)
        
        # Get samples from first epoch
        samples_epoch1 = [dataset[i] for i in range(len(dataset))]
        
        # Get samples from second epoch
        samples_epoch2 = [dataset[i] for i in range(len(dataset))]
        
        # Should be identical
        for s1, s2 in zip(samples_epoch1, samples_epoch2):
            assert torch.equal(s1["boxes"], s2["boxes"])
            assert torch.equal(s1["labels"], s2["labels"])

    def test_memory_cleanup(self, temp_coco_dataset):
        """Test that dataset can be properly cleaned up."""
        img_folder, ann_file = temp_coco_dataset
        dataset = ElephantCocoDataset(img_folder, ann_file)
        
        # Access some samples
        for i in range(len(dataset)):
            _ = dataset[i]
        
        # Delete and cleanup
        del dataset
        gc.collect()

    def test_boxes_labels_correspondence(self, temp_coco_dataset):
        """Test that boxes and labels have matching lengths."""
        img_folder, ann_file = temp_coco_dataset
        dataset = ElephantCocoDataset(img_folder, ann_file)
        
        for i in range(len(dataset)):
            sample = dataset[i]
            assert len(sample["boxes"]) == len(sample["labels"])

    def test_all_images_accessible(self, temp_coco_dataset):
        """Test that all images can be loaded without errors."""
        img_folder, ann_file = temp_coco_dataset
        dataset = ElephantCocoDataset(img_folder, ann_file)
        
        for i in range(len(dataset)):
            sample = dataset[i]
            # Check that image was loaded
            assert sample["image"].numel() > 0
            assert torch.isfinite(sample["image"]).all()


class TestElephantCocoDatasetBoundaryConditions:
    """Test boundary conditions and special cases."""

    def test_single_pixel_box(self):
        """Test handling of very small boxes."""
        with tempfile.TemporaryDirectory() as tmpdir:
            tmpdir = Path(tmpdir)
            img_dir = tmpdir / "images"
            img_dir.mkdir()
            
            # Create image
            img = Image.new('RGB', (640, 480))
            img.save(img_dir / "image_0.jpg")
            
            # Box with width=1, height=1
            annotations = {
                "images": [
                    {"id": 0, "file_name": "image_0.jpg", "width": 640, "height": 480},
                ],
                "annotations": [
                    {"id": 0, "image_id": 0, "category_id": 1, "bbox": [100, 100, 1, 1], "area": 1, "iscrowd": 0},
                ],
                "categories": [{"id": 1, "name": "elephant"}]
            }
            
            ann_file = tmpdir / "annotations.json"
            with open(ann_file, 'w') as f:
                json.dump(annotations, f)
            
            dataset = ElephantCocoDataset(str(img_dir), str(ann_file))
            sample = dataset[0]
            
            # Should accept 1x1 box as valid
            assert len(sample["boxes"]) == 1

    def test_large_box(self):
        """Test handling of large boxes spanning entire image."""
        with tempfile.TemporaryDirectory() as tmpdir:
            tmpdir = Path(tmpdir)
            img_dir = tmpdir / "images"
            img_dir.mkdir()
            
            img = Image.new('RGB', (640, 480))
            img.save(img_dir / "image_0.jpg")
            
            # Box covering entire image
            annotations = {
                "images": [
                    {"id": 0, "file_name": "image_0.jpg", "width": 640, "height": 480},
                ],
                "annotations": [
                    {"id": 0, "image_id": 0, "category_id": 1, "bbox": [0, 0, 640, 480], "area": 307200, "iscrowd": 0},
                ],
                "categories": [{"id": 1, "name": "elephant"}]
            }
            
            ann_file = tmpdir / "annotations.json"
            with open(ann_file, 'w') as f:
                json.dump(annotations, f)
            
            dataset = ElephantCocoDataset(str(img_dir), str(ann_file))
            sample = dataset[0]
            
            assert len(sample["boxes"]) == 1
            box = sample["boxes"][0]
            assert box[0] == 0
            assert box[1] == 0
            assert box[2] == 640
            assert box[3] == 480


if __name__ == "__main__":
    pytest.main([__file__, "-v"])



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

"""Substation segmentation dataset."""

import glob
import os
from collections.abc import Callable, Sequence

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import torch
from matplotlib.figure import Figure
from torch import Tensor

from torchgeo.datasets.errors import DatasetNotFoundError
from torchgeo.datasets.geo import NonGeoDataset
from torchgeo.datasets.utils import Path, download_url, extract_archive

from typing import Any, ClassVar
from matplotlib import patches
from PIL import Image

import pycocotools.coco
import pycocotools
import pdb
import skimage

import requests


def download_file_from_presigned(url, target_folder, filename):

    target_path = os.path.join(target_folder, filename)
    if os.path.isfile(target_path):
        print(f"File already exists at: {target_path}. Skipping download.")
    else:
        # Stream download for large files
        with requests.get(url, stream=True) as r:
            r.raise_for_status()
            with open(target_path, 'wb') as f:
                for chunk in r.iter_content(chunk_size=8192):
                    if chunk:
                        f.write(chunk)


def convert_coco_poly_to_mask(
    segmentations: list[int], height: int, width: int
) -> Tensor:
    """Convert coco polygons to mask tensor.

    Args:
        segmentations (List[int]): polygon coordinates
        height (int): image height
        width (int): image width

    Returns:
        Tensor: Mask tensor

    Raises:
        DependencyNotFoundError: If pycocotools is not installed.
    """
    masks = []
    for polygons in segmentations:
        rles = pycocotools.mask.frPyObjects(polygons, height, width)
        mask = pycocotools.mask.decode(rles)
        mask = torch.as_tensor(mask, dtype=torch.uint8)
        mask = mask.any(dim=2)
        masks.append(mask)
    masks_tensor = torch.stack(masks, dim=0)
    return masks_tensor


class ConvertCocoAnnotations:
    """Callable for converting the boxes, masks and labels into tensors.

    This is a modified version of ConvertCocoPolysToMask() from torchvision found in
    https://github.com/pytorch/vision/blob/v0.14.0/references/detection/coco_utils.py
    """

    def __call__(self, sample: dict[str, Any]) -> dict[str, Any]:
        """Converts MS COCO fields (boxes, masks & labels) from list of ints to tensors.

        Args:
            sample: Sample

        Returns:
            Processed sample
        """
        image = sample['image']
        h, w = image.size()[-2:]        
        target = sample['label']

        image_id = target['image_id']
        image_id = torch.tensor([image_id])

        anno = target['annotations']

        anno = [obj for obj in anno if obj['iscrowd'] == 0]

        bboxes = [obj['bbox'] for obj in anno]
        # guard against no boxes via resizing
        boxes = torch.as_tensor(bboxes, dtype=torch.float32).reshape(-1, 4)
        boxes[:, 2:] += boxes[:, :2]
        boxes[:, 0::2].clamp_(min=0, max=w)
        boxes[:, 1::2].clamp_(min=0, max=h)

        categories = [obj['category_id'] for obj in anno]
        classes = torch.tensor(categories, dtype=torch.int64)

        segmentations = [obj['segmentation'] for obj in anno]

        masks = convert_coco_poly_to_mask(segmentations, h, w)

        keep = (boxes[:, 3] > boxes[:, 1]) & (boxes[:, 2] > boxes[:, 0])
        boxes = boxes[keep]
        classes = classes[keep]

        target = {'boxes': boxes, 'labels': classes, 'image_id': image_id}
        if masks.nelement() > 0:
            masks = masks[keep]
            target['masks'] = masks

        # for conversion to coco api
        area = torch.tensor([obj['area'] for obj in anno])
        iscrowd = torch.tensor([obj['iscrowd'] for obj in anno])
        target['area'] = area
        target['iscrowd'] = iscrowd
        return {'image': image, 'label': target}


class Substation(NonGeoDataset):
    """Substation dataset.

    The `Substation <https://github.com/Lindsay-Lab/substation-seg>`__
    dataset is curated by TransitionZero and sourced from publicly
    available data repositories, including OpenSreetMap (OSM) and
    Copernicus Sentinel data. The dataset consists of Sentinel-2
    images from 27k+ locations; the task is to segment power-substations,
    which appear in the majority of locations in the dataset.
    Most locations have 4-5 images taken at different timepoints
    (i.e., revisits).

    Dataset Format:

    * .npz file for each datapoint

    Dataset Features:

    * 26,522 image-mask pairs stored as numpy files.
    * Data from 5 revisits for most locations.
    * Multi-temporal, multi-spectral images (13 channels) paired with masks,
      with a spatial resolution of 228x228 pixels

    If you use this dataset in your research, please cite the following:

    * https://doi.org/10.48550/arXiv.2409.17363
    """

    directory = 'Substation'
    filename_images = 'image_stack.tar.gz'
    filename_masks = 'mask.tar.gz'
    filename_detection_labels = 'annotations.json'
    url_for_images = 'https://storage.googleapis.com/tz-ml-public/substation-over-10km2-csv-main-444e360fd2b6444b9018d509d0e4f36e/image_stack.tar.gz'
    url_for_masks = 'https://storage.googleapis.com/tz-ml-public/substation-over-10km2-csv-main-444e360fd2b6444b9018d509d0e4f36e/mask.tar.gz'
    url_for_detection_labels = 'annotations.json'
    categories = ( 'background', 'substation')

    def __init__(
        self,
        root: Path,
        bands: Sequence[int],
        mask_2d: bool,
        timepoint_aggregation: str = 'concat',
        download: bool = False,
        checksum: bool = False,
        transforms: Callable[[dict[str, Tensor]], dict[str, Tensor]] | None = None,
        num_of_timepoints: int = 4,
        use_timepoints: bool = False,
        mode: str = 'segmentation',
        split: str = 'train',
        dataset_version: str = 'full',
        plot_indexes: Sequence[int] = [2,1,0]
    ) -> None:
        """Initialize the Substation.

        Args:
            root: Path to the directory containing the dataset.
            bands: Channels to use from the image.
            mask_2d: Whether to use a 2D mask.
            timepoint_aggregation: How to aggregate multiple timepoints.
            download: Whether to download the dataset if it is not found.
            checksum: Whether to verify the dataset after downloading.
            transforms: A transform takes input sample and returns a transformed version.
            num_of_timepoints: Number of timepoints to use for each image.
            use_timepoints: Whether to use multiple timepoints for each image.
            mode: Either segmentation or object_detection.
            split: Split to load (either train/val/test).
            dataset_version: Dataset version to select (either'full' or 'geobench') for '' for full dataset or 'substation_meta_splits_geobench.csv' for GEO-Bench subsampled dataset.
            plot_indexes: list of indexes to use for plotting.
        """
        
        self.root = root
        self.bands = bands
        self.mask_2d = mask_2d
        self.timepoint_aggregation = timepoint_aggregation
        self.download = download
        self.use_timepoints = use_timepoints
        self.checksum = checksum
        self.transforms = transforms
        self.image_dir = os.path.join(root, 'image_stack')
        self.mode = mode
        if self.mode == 'segmentation':
            self.mask_dir = os.path.join(root, 'mask')
        self.num_of_timepoints = num_of_timepoints
        self._verify()
        self.image_filenames = pd.Series(sorted(os.listdir(self.image_dir)))

        self.split = split
        self.dataset_version = dataset_version
        if self.dataset_version == 'full':

            self.meta_df = pd.read_csv(os.path.join(self.root, 'substation_meta_splits_full.csv'))

        elif self.dataset_version == 'geobench':

            self.meta_df = pd.read_csv(os.path.join(self.root, 'substation_meta_splits_geobench.csv'))

        self.meta_df = self.meta_df[self.meta_df['split'].values == self.split]
        self.meta_df['image'] = [x.split('/')[-1] for x in self.meta_df['image'].values]

        if self.mode == 'object_detection':
            self.coco = pycocotools.coco.COCO(os.path.join(self.root, self.filename_detection_labels))
            self.coco_convert = ConvertCocoAnnotations()

        self.image_filenames = self.meta_df['image'].values
        
        assert self.timepoint_aggregation in {'first', 'last', 'random', 'concat', 'median', 'identity'}
        if self.use_timepoints == False:
            assert self.timepoint_aggregation in {'first', 'last', 'random'}

        self.plot_indexes = plot_indexes

    def _load_image(self, image_path):
        
        image = np.load(image_path)['arr_0']
        
        # selecting channels
        image = image[:, self.bands, :, :]
        # handling multiple images across timepoints
        
        if self.use_timepoints:
            if image.shape[0] < self.num_of_timepoints:

                # Padding: cycle through existing timepoints
                padded_images = []
                for i in range(self.num_of_timepoints):
                    padded_images.append(image[i % image.shape[0]])
                image = np.stack(padded_images)
            elif image.shape[0] > self.num_of_timepoints:
                # Removal: take the most recent timepoints
                image = image[-self.num_of_timepoints :]

            if self.timepoint_aggregation == 'concat':
                image = np.reshape(
                    image, (-1, image.shape[2], image.shape[3])
                )  # (num_of_timepoints*channels,h,w)
            elif self.timepoint_aggregation == 'median':
                image = np.median(image, axis=0)
            elif self.timepoint_aggregation == 'identity':
                ## No changes
                image = image
        else:
            if self.timepoint_aggregation == 'first':
                image = image[0]
            elif self.timepoint_aggregation == 'last':
                image = image[-1]
            elif self.timepoint_aggregation == 'random':
                image = image[np.random.randint(image.shape[0])]
            
        tensor_image = torch.from_numpy(image)
        tensor_image = tensor_image.float()

        return tensor_image

    def _load_od_target(self, id_):
        """Load the annotations for a single image.

        Args:
            id_: unique ID of the image

        Returns:
            the annotations
        """
        # Images in the "negative" image set have no annotations
        annot = []
        ann_ids = self.coco.getAnnIds(imgIds=self.meta_df['id'].values[id_])
        annot = self.coco.loadAnns(ann_ids)

        target = dict(image_id=id_, annotations=annot)

        return target
    
    def _load_segmentation_mask(self, mask_path):
    
        mask = np.load(mask_path)['arr_0']
        mask[mask != 3] = 0
        mask[mask == 3] = 1
        mask = torch.from_numpy(mask).long()
        mask = mask.unsqueeze(dim=0)

        if self.mask_2d:
            mask_0 = 1.0 - mask
            mask = torch.concat([mask_0, mask], dim=0)
        mask = mask.squeeze()
        
        return mask


    def __getitem__(self, index: int) -> dict[str, Tensor]:
        """Get an item from the dataset by index.

        Args:
            index: Index of the item to retrieve.

        Returns:
            A dictionary containing the image and corresponding mask.
        """
        image_filename = self.image_filenames[index]
        image_path = os.path.join(self.image_dir, image_filename)
        image = self._load_image(image_path)

        if self.mode == "segmentation":
            mask_path = os.path.join(self.mask_dir, image_filename)
            mask = self._load_segmentation_mask(mask_path)
            sample = {'image': image, 'mask': mask}

        if self.mode == "object_detection":
            
            od_label = self._load_od_target(index)
            sample: dict[str, Any] = {'image': image, 'label': od_label}
            
            if sample['label']['annotations']:
                sample = self.coco_convert(sample)
                sample['class'] = sample['label']['labels']
                sample['boxes'] = sample['label']['boxes']
                sample['masks'] = sample['label']['masks'].float()
                sample['labels'] = sample.pop('class')
                
        if self.transforms is not None:
            
            sample = self.transforms(sample)

        return sample

    def __len__(self) -> int:
        """Returns the number of items in the dataset."""
        return len(self.image_filenames)
    
    def _plot_segmentation(self,
                           sample: dict[str, Tensor],
                           show_titles: bool = True,
                           suptitle: str | None = None) -> Figure:
        """Plot a sample from the dataset.

        Args:
            sample: a sample returned by :meth:`__getitem__`
            show_titles: flag indicating whether to show titles above each panel
            suptitle: optional string to use as a suptitle

        Returns:
            A matplotlib Figure containing the rendered sample.
        """
        ncols = 2

        image = sample['image']
        image = image[-1] if len(image.shape) == 4 else image
        # get rgb indexes from 
        image = image[ self.plot_indexes, :, :].permute(1, 2, 0).numpy()
        
        if image.mean() > 1:
            image = image / 10000

        image = np.clip(image, 0, 1)

        if self.mask_2d:
            mask = sample['mask'][0].squeeze(0).cpu().numpy()
        else:
            mask = sample['mask'].cpu().numpy()
        showing_predictions = 'prediction' in sample
        if showing_predictions:
            prediction = sample['prediction'].cpu().numpy()
            if self.mask_2d:
                prediction = prediction[0]
            ncols = 3

        fig, axs = plt.subplots(ncols=ncols, figsize=(4 * ncols, 4))
        axs[0].imshow(image)
        axs[0].axis('off')
        axs[1].imshow(mask, cmap='gray', interpolation='none')
        axs[1].axis('off')

        if show_titles:
            axs[0].set_title('Image')
            axs[1].set_title('Mask')

        if showing_predictions:
            axs[2].imshow(prediction, cmap='gray', interpolation='none')
            axs[2].axis('off')
            if show_titles:
                axs[2].set_title('Prediction')

        if suptitle:
            fig.suptitle(suptitle)

        return fig
    
    def _plot_object_detection(self,
                               sample: dict[str, Tensor],
                               show_titles: bool = True,
                               suptitle: str | None = None,
                               show_feats: str | None = 'both',
                               box_alpha: float = 0.7,
                               mask_alpha: float = 0.7,
                               confidence_score = 0.5) -> Figure:

        """Plot a sample from the dataset.

        Args:
            sample: a sample returned by :meth:`__getitem__`
            suptitle: optional string to use as a suptitle
            show_titles: flag indicating whether to show titles above each panel
            show_feats: optional string to pick features to be shown: boxes, masks, both
            box_alpha: alpha value of box
            mask_alpha: alpha value of mask
            confidence_score: 
            
        Returns:
            a matplotlib Figure with the rendered sample

        Raises:
            AssertionError: if ``show_feats`` argument is invalid
            DependencyNotFoundError: If plotting masks and scikit-image is not installed.

        .. versionadded:: 0.4
        """
        assert show_feats in {'boxes', 'masks', 'both'}
        
        image = sample['image']
        image = image[-1] if len(image.shape) == 4 else image
        # get indexes to plot 
        image = image[ self.plot_indexes, :, :].permute(1, 2, 0).numpy()
        
        if image.mean() > 1:
            image = image / 10000

        image = np.clip(image, 0, 1)

        boxes = sample['boxes'].cpu().numpy()
        labels = sample['labels'].cpu().numpy()

        if 'masks' in sample:
            masks = [mask.squeeze().cpu().numpy() for mask in sample['masks']]

        n_gt = len(boxes)

        ncols = 1
        show_predictions = 'prediction_labels' in sample

        if show_predictions:
            show_pred_boxes = False
            show_pred_masks = False
            prediction_labels = sample['prediction_labels'].numpy()
            prediction_scores = sample['prediction_scores'].numpy()
            if 'prediction_boxes' in sample:
                prediction_boxes = sample['prediction_boxes'].numpy()
                show_pred_boxes = True
            if 'prediction_masks' in sample:
                prediction_masks = sample['prediction_masks'].numpy()
                show_pred_masks = True

            n_pred = len(prediction_labels)
            ncols += 1

        # Display image
        fig, axs = plt.subplots(ncols=ncols, squeeze=False, figsize=(ncols * 10, 13))
        axs[0, 0].imshow(image)
        axs[0, 0].axis('off')

        cm = plt.get_cmap('gist_rainbow')
        for i in range(n_gt):
            class_num = labels[i]
            color = cm(class_num / len(self.categories))

            # Add bounding boxes
            x1, y1, x2, y2 = boxes[i]
            if show_feats in {'boxes', 'both'}:
                r = patches.Rectangle(
                    (x1, y1),
                    x2 - x1,
                    y2 - y1,
                    linewidth=2,
                    alpha=box_alpha,
                    linestyle='dashed',
                    edgecolor=color,
                    facecolor='none',
                )
                axs[0, 0].add_patch(r)

            # Add labels
            label = self.categories[class_num]
            caption = label
            axs[0, 0].text(
                x1, y1 - 8, caption, color='white', size=11, backgroundcolor='none'
            )

            # Add masks
            if show_feats in {'masks', 'both'} and 'masks' in sample:
                mask = masks[i]
                contours = skimage.measure.find_contours(mask, 0.5)
                for verts in contours:
                    verts = np.fliplr(verts)
                    p = patches.Polygon(
                        verts, facecolor=color, alpha=mask_alpha, edgecolor='white'
                    )
                    axs[0, 0].add_patch(p)

            if show_titles:
                axs[0, 0].set_title('Ground Truth')
        if show_predictions:
            axs[0, 1].imshow(image)
            axs[0, 1].axis('off')
            for i in range(n_pred):
                score = prediction_scores[i]
                if score < confidence_score:
                    continue

                class_num = prediction_labels[i]
                color = cm(class_num / len(self.categories))

                if show_pred_boxes:
                    # Add bounding boxes
                    x1, y1, x2, y2 = prediction_boxes[i]
                    r = patches.Rectangle(
                        (x1, y1),
                        x2 - x1,
                        y2 - y1,
                        linewidth=2,
                        alpha=box_alpha,
                        linestyle='dashed',
                        edgecolor=color,
                        facecolor='none',
                    )
                    axs[0, 1].add_patch(r)

                    # Add labels
                    label = self.categories[class_num]
                    caption = f'{label} {score:.3f}'
                    axs[0, 1].text(
                        x1,
                        y1 - 8,
                        caption,
                        color='white',
                        size=11,
                        backgroundcolor='none',
                    )

                # Add masks
                if show_pred_masks:

                    mask = prediction_masks[i][0]
                    contours = skimage.measure.find_contours(mask, 0.5)
                    for verts in contours:
                        verts = np.fliplr(verts)
                        p = patches.Polygon(
                            verts, facecolor=color, alpha=mask_alpha, edgecolor='white'
                        )
                        axs[0, 1].add_patch(p)

            if show_titles:
                axs[0, 1].set_title('Prediction')
        if suptitle is not None:
            plt.suptitle(suptitle)

        plt.tight_layout()

        return fig
 
    def plot(
        self,
        sample: dict[str, Tensor],
        show_titles: bool = True,
        suptitle: str | None = None,
        **kwargs) -> Figure:

        if self.mode == 'segmentation':
            figure = self._plot_segmentation(sample, show_titles, suptitle)
        else:
            figure = self._plot_object_detection(sample, show_titles, suptitle, **kwargs)

        return figure

    def _extract(self) -> None:
        """Extract the dataset."""
        img_pathname = os.path.join(self.root, self.filename_images)
        extract_archive(img_pathname)

        if self.mode == "segmentation":
            mask_pathname = os.path.join(self.root, self.filename_masks)
            extract_archive(mask_pathname)

    def _verify(self) -> None:
        """Verify the integrity of the dataset."""
        # Check if the extracted files already exist
        image_path = os.path.join(self.image_dir, '*.npz')

        if self.mode == "segmentation":
            mask_path = os.path.join(self.mask_dir, '*.npz')
            if glob.glob(image_path) and glob.glob(mask_path):
                return
        elif self.mode == "object_detection":
            detection_labels_exist = os.path.exists(os.path.join(self.root, self.filename_detection_labels))
            if glob.glob(image_path) and detection_labels_exist:
                return

        # Check if the tar.gz files for images and masks have already been downloaded
        image_exists = os.path.exists(os.path.join(self.root, self.filename_images))

        if self.mode == "segmentation":
            mask_exists = os.path.exists(os.path.join(self.root, self.filename_masks))
            if image_exists and mask_exists:
                self._extract()
                return
        elif self.mode == "object_detection":
            if image_exists:
                self._extract()
                return

        # If dataset files are missing and download is not allowed, raise an error
        if not getattr(self, 'download', True):
            raise DatasetNotFoundError(self)

        # Download and extract the dataset
        self._download()
        self._extract()

    def _download(self) -> None:
        """Download the dataset and extract it."""
        # Download and verify images
        download_url(
            self.url_for_images,
            self.root,
            filename=self.filename_images,
            md5='bc0e3d2565f30b1c84a0d4cf37d44be6' if self.checksum else None,
        )
        extract_archive(os.path.join(self.root, self.filename_images), self.root)

        # Download splits
        download_file_from_presigned("https://ibm.box.com/shared/static/1wv1fiva5w722aka3r8dd0vrmly4zat0.csv", 
                                     self.root, 
                                     "substation_meta_splits_geobench.csv")
        download_file_from_presigned("https://ibm.box.com/shared/static/sgfsyewhtjpwvigu9wrsc8ia26t1ug9d.csv", 
                                     self.root, 
                                     "substation_meta_splits_full.csv")

        if self.mode == 'segmentation':
            # Download and verify masks
            download_url(
                self.url_for_masks,
                self.root,
                filename=self.filename_masks,
                md5='919bb9599f47f44f17a1a4ecce56d81c' if self.checksum else None,
            )
            extract_archive(os.path.join(self.root, self.filename_masks), self.root)

        elif self.mode == 'object_detection':

            download_file_from_presigned("https://ibm.box.com/shared/static/sulkq5q5v3gthjf7o2jrpyn9kbw1uszd.json", 
                                         self.root, 
                                         "annotations.json")

